import logging
from typing import cast

from housekeeper.store.models import Version

from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG
from cg.constants.scout import (
    NALLO_CASE_TAGS,
    NALLO_RANK_MODEL_THRESHOLD,
    NALLO_RANK_MODEL_VERSION_SNV,
    NALLO_RANK_MODEL_VERSION_SV,
    NALLO_SAMPLE_TAGS,
    GenomeBuild,
    UploadTrack,
)
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.meta.workflow.nallo import NalloAnalysisAPI
from cg.models.scout.scout_load_config import NalloLoadConfig, ScoutNalloIndividual
from cg.store.models import Analysis

LOG = logging.getLogger(__name__)


class NalloConfigBuilder(ScoutConfigBuilder):
    def __init__(
        self,
        nallo_analysis_api: NalloAnalysisAPI,
        lims_api: LimsAPI,
        madeline_api: MadelineAPI,
    ):
        super().__init__(
            lims_api=lims_api,
        )
        self.case_tags = CaseTags(**NALLO_CASE_TAGS)
        self.sample_tags = SampleTags(**NALLO_SAMPLE_TAGS)
        self.nallo_analysis_api: NalloAnalysisAPI = nallo_analysis_api
        self.lims_api: LimsAPI = lims_api
        self.madeline_api: MadelineAPI = madeline_api

    def build_load_config(self, hk_version: Version, analysis: Analysis) -> NalloLoadConfig:
        """Create a NALLO specific load config for uploading analysis to Scout."""
        LOG.info("Build load config for NALLO case")
        load_config = NalloLoadConfig(
            track=UploadTrack.RARE_DISEASE.value,
            delivery_report=self.get_file_from_hk(
                hk_tags={HK_DELIVERY_REPORT_TAG}, hk_version=hk_version
            ),
        )
        self.add_common_info_to_load_config(load_config=load_config, analysis=analysis)
        load_config.gene_panels = self.nallo_analysis_api.get_aggregated_panels(
            customer_id=analysis.case.customer.internal_id,
            default_panels=set(analysis.case.panels),
        )
        self.include_case_files(load_config=load_config, hk_version=hk_version)
        LOG.info("Building samples")
        self.get_sample_information(
            load_config=load_config, analysis=analysis, hk_version=hk_version
        )
        self.include_pedigree_picture(load_config=load_config, analysis=analysis)
        load_config.human_genome_build = GenomeBuild.hg38
        load_config.rank_score_threshold = NALLO_RANK_MODEL_THRESHOLD
        load_config.rank_model_version = NALLO_RANK_MODEL_VERSION_SNV
        load_config.sv_rank_model_version = NALLO_RANK_MODEL_VERSION_SV
        return load_config

    def include_case_files(self, load_config: NalloLoadConfig, hk_version: Version = None) -> None:
        """Include case level files for NALLO case."""
        LOG.info("Including NALLO specific case level files")
        for scout_key in NALLO_CASE_TAGS.keys():
            self._include_case_file(
                load_config=load_config, scout_key=scout_key, hk_version=hk_version
            )

    def _include_case_file(
        self, load_config: NalloLoadConfig, scout_key: str, hk_version: Version
    ) -> None:
        """Include the file path associated to a scout configuration parameter if the corresponding housekeeper tags
        are found. Otherwise return None."""
        file_path = self.get_file_from_hk(
            hk_tags=getattr(self.case_tags, scout_key), hk_version=hk_version
        )
        setattr(load_config, scout_key, file_path)

    def include_sample_files(
        self, config_sample: ScoutNalloIndividual, hk_version: Version
    ) -> None:
        """Include sample level files that are optional."""
        LOG.info("Including NALLO specific sample level files")
        sample_id: str = cast(str, config_sample.sample_id)
        config_sample.chromograph_images.autozygous = self.remove_chromosome_substring(
            self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_autozyg,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.chromograph_images.coverage = self.remove_chromosome_substring(
            self.get_sample_file(
                hk_tags=self.sample_tags.chromograph_coverage,
                sample_id=sample_id,
                hk_version=hk_version,
            )
        )
        config_sample.d4_file = self.get_sample_file(
            hk_tags=cast(set[str], self.sample_tags.d4_file),
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.paraphase_alignment_path = self.get_sample_file(
            hk_tags=cast(set[str], self.sample_tags.paraphase_alignment_path),
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.tiddit_coverage_wig = self.get_sample_file(
            hk_tags=cast(set[str], self.sample_tags.hificnv_coverage),
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.minor_allele_frequency_wig = self.get_sample_file(
            hk_tags=cast(set[str], self.sample_tags.minor_allele_frequency_wig),
            sample_id=sample_id,
            hk_version=hk_version,
        )
        alignment_and_mt_bam_path: str | None = self.get_sample_file(
            hk_tags=cast(set[str], self.sample_tags.alignment_path),
            sample_id=sample_id,
            hk_version=hk_version,
        )
        config_sample.alignment_path = alignment_and_mt_bam_path
        config_sample.mt_bam = alignment_and_mt_bam_path
        config_sample.phase_blocks = self.get_sample_file(
            hk_tags=cast(set[str], self.sample_tags.phase_blocks),
            sample_id=sample_id,
            hk_version=hk_version,
        )
        self.include_reviewer_files(config_sample=config_sample, hk_version=hk_version)
        config_sample.reviewer.trgt = (
            True
            if any(
                [
                    config_sample.reviewer.alignment,
                    config_sample.reviewer.alignment_index,
                    config_sample.reviewer.vcf,
                    config_sample.reviewer.catalog,
                ]
            )
            else None
        )



from abc import ABC, abstractmethod
from typing import TypeVar

from cg.services.analysis_starter.configurator.abstract_model import CaseConfig

SpecificCaseConfig = TypeVar("SpecificCaseConfig", bound=CaseConfig)


class Configurator(ABC):
    @abstractmethod
    def configure(self, case_id: str, **flags) -> CaseConfig:
        """Abstract method to configure a case."""
        pass

    @abstractmethod
    def get_config(self, case_id: str, **flags) -> CaseConfig:
        """Abstract method to get the (pre-existing) config for a case."""
        pass

    @staticmethod
    def _set_flags(config: SpecificCaseConfig, **flags) -> SpecificCaseConfig:
        curated_flags: dict = {key: value for key, value in flags.items() if value is not None}
        return config.model_validate(config.model_dump(by_alias=True) | curated_flags)

    @abstractmethod
    def _ensure_required_config_files_exist(self, **kwargs) -> None:
        pass



import logging

from cg.constants.devices import DeviceType
from cg.constants.priority import Priority
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.exc import ApplicationDoesNotHaveHiFiYieldError
from cg.store.models import Case, Sample

LOG = logging.getLogger(__name__)


def case_pass_sequencing_qc_on_reads(case: Case) -> bool:
    """
    Get the sequencing QC of a case. The checks are performed in the following order:
    1. If the case is a ready-made library, the ready made library QC is used.
    2. If the case is express priority, the express QC is used.
    3. If neither of the above conditions are met, it checks if all samples have enough reads.

    The order of these checks is important because the ready-made library QC is more relaxed,
    and should be checked before the express QC, and the express QC should be checked before the
    standard QC.
    """
    if is_case_ready_made_library(case):
        return ready_made_library_case_pass_sequencing_qc(case)
    if is_case_express_priority(case):
        return express_case_pass_sequencing_qc_on_reads(case)
    return all(sample_has_enough_reads(sample) for sample in case.samples)


def case_pass_sequencing_qc_on_hifi_yield(case: Case) -> bool:
    """
    Get the sequencing QC of a case using yield. The checks are performed in the following order:
    1. If the case is express priority, the express QC is used.
    2. If the above condition is not met then it checks if all samples have enough yield.

    The express QC should be checked before the standard QC.
    """
    if is_case_express_priority(case):
        return express_case_pass_sequencing_qc_on_hifi_yield(case)
    return all(sample_has_enough_hifi_yield(sample) for sample in case.samples)


def express_case_pass_sequencing_qc_on_reads(case: Case) -> bool:
    """
    Checks if all samples in an express case have enough reads.
    """
    return all(express_sample_has_enough_reads(sample) for sample in case.samples)


def express_case_pass_sequencing_qc_on_hifi_yield(case: Case) -> bool:
    """
    Checks if all samples in an express case have enough hifi yield.
    """
    return all(express_sample_has_enough_yield(sample) for sample in case.samples)


def express_sample_pass_sequencing_qc_on_reads(sample: Sample) -> bool:
    return express_sample_has_enough_reads(sample)


def express_sample_has_enough_reads(sample: Sample) -> bool:
    """
    Checks if given express sample has enough reads. Gets the threshold from the sample's
    application version.
    """
    express_reads_threshold: int = get_express_reads_threshold_for_sample(sample)
    enough_reads: bool = sample.reads >= express_reads_threshold
    if not enough_reads:
        LOG.warning(f"Sample {sample.internal_id} has too few reads.")
    return enough_reads


def express_sample_has_enough_yield(sample: Sample) -> bool:
    if not sample.hifi_yield:
        LOG.debug(f"Sample {sample.internal_id} has no hifi yield.")
        return False

    express_yield_threshold: int = get_express_yield_threshold_for_sample(sample)
    enough_yield: bool = sample.hifi_yield >= express_yield_threshold
    if not enough_yield:
        LOG.warning(f"Sample {sample.internal_id} does not have enough yield.")
    return enough_yield


def sample_pass_sequencing_qc_on_reads(sample: Sample) -> bool:
    """
    Get the standard sequencing QC of a sample. The checks are performed in the following order:
    1. If the sample is express priority, the express QC is used.
    2. If the sample is a ready-made library, the ready made library QC is used.
    3. If neither of the above conditions are met, it checks if the sample has enough reads.

    The order of these checks is important because the express QC is more relaxed, and should be
    checked before the standard QC.
    """
    if is_sample_express_priority(sample):
        return express_sample_pass_sequencing_qc_on_reads(sample)
    if is_sample_ready_made_library(sample):
        return ready_made_library_sample_has_enough_reads(sample)
    return sample_has_enough_reads(sample)


def ready_made_library_case_pass_sequencing_qc(case: Case) -> bool:
    """
    Checks if the sum of reads of all samples in a ready made library case is enough.

    The threshold is the expected reads for any sample in the case.
    """
    threshold: int = case.samples[0].expected_reads_for_sample
    reads: int = sum(sample.reads for sample in case.samples)
    enough_reads: bool = reads >= threshold
    if not enough_reads:
        LOG.warning(f"Case {case.internal_id} has too few reads.")
    return enough_reads


def all_samples_in_case_have_reads(case: Case) -> bool:
    """
    Check if all samples have reads.
    """
    passed_quality_check: bool = all(sample.has_reads for sample in case.samples)
    if not passed_quality_check:
        LOG.warning("Not all samples in case have reads.")
    return passed_quality_check


def any_sample_in_case_has_reads(case: Case) -> bool:
    """
    Check if any sample has reads.
    """
    passed_quality_check: bool = any(sample.has_reads for sample in case.samples)
    if not passed_quality_check:
        LOG.warning("No samples in case have reads.")
    return passed_quality_check


def raw_data_case_pass_qc(case: Case) -> bool:
    if is_case_ready_made_library(case):
        return ready_made_library_case_pass_sequencing_qc(case)
    if is_first_sample_yield_based_and_processed(case):
        return all(sample_has_enough_hifi_yield(sample) for sample in case.samples)
    elif is_first_sample_reads_based_and_processed(case):
        return all(sample_has_enough_reads(sample) for sample in case.samples)
    LOG.warning(f"Not all samples for case {case.internal_id} have been post-processed.")
    return False


def is_first_sample_yield_based_and_processed(case: Case) -> bool:
    sample: Sample = case.samples[0]
    if metrics := sample.sample_run_metrics:
        return metrics[0].type == DeviceType.PACBIO
    return False


def is_first_sample_reads_based_and_processed(case: Case) -> bool:
    sample: Sample = case.samples[0]
    if metrics := sample.sample_run_metrics:
        return metrics[0].type == DeviceType.ILLUMINA
    return False


def is_case_express_priority(case: Case) -> bool:
    """
    Check if a case is express priority.
    """
    return case.priority == Priority.express


def get_express_reads_threshold_for_sample(sample: Sample) -> int:
    """
    Get the express reads threshold for a sample.
    """
    return round(sample.application_version.application.target_reads / 2)


def get_express_yield_threshold_for_sample(sample: Sample) -> int:
    if threshold := sample.application_version.application.expected_express_hifi_yield:
        return threshold
    else:
        raise ApplicationDoesNotHaveHiFiYieldError(
            f"Application for sample {sample.internal_id} does not have target HiFi yield."
        )


def is_sample_ready_made_library(sample: Sample) -> bool:
    return sample.prep_category == SeqLibraryPrepCategory.READY_MADE_LIBRARY


def is_case_ready_made_library(case: Case) -> bool:
    """
    Check if all samples in case are ready made libraries.
    """
    return all(is_sample_ready_made_library(sample) for sample in case.samples)


def ready_made_library_sample_has_enough_reads(sample: Sample) -> bool:
    """
    Check if a given sample from a ready made library has enough reads.
    """
    if not sample.has_reads:
        LOG.warning(f"Sample {sample.internal_id} has no reads.")
    return sample.has_reads


def sample_has_enough_reads(sample: Sample) -> bool:
    """
    Check if the sample has more or equal reads than the expected reads for the sample.
    """
    enough_reads: bool = sample.reads >= sample.expected_reads_for_sample
    if not enough_reads:
        LOG.warning(f"Sample {sample.internal_id} has too few reads.")
    return enough_reads


def sample_has_enough_hifi_yield(sample: Sample) -> bool:
    """
    Return true if the sample's HiFi yield is greater than or equal to the threshold.
    Returns false if the HiFi yield is lower than the threshold or None.
    Raises:
        ApplicationDoesNotHaveHiFiYieldError if the sample doesn't have expected HiFi yield.
    """
    if not sample.expected_hifi_yield:
        raise ApplicationDoesNotHaveHiFiYieldError(
            f"Application for sample {sample.internal_id} does not have target HiFi yield."
        )

    if not sample.hifi_yield:
        LOG.debug(f"Sample {sample.internal_id} has no hifi yield.")
        return False

    enough_hifi_yield: bool = sample.hifi_yield >= sample.expected_hifi_yield
    if not enough_hifi_yield:
        LOG.warning(f"Sample {sample.internal_id} does not have enough HiFi yield.")
    return enough_hifi_yield


def is_sample_express_priority(sample: Sample) -> bool:
    """
    Check if a sample is express priority.
    """
    return sample.priority == Priority.express



import logging
from pathlib import Path

from cg.apps.lims import LimsAPI
from cg.constants.constants import DEFAULT_CAPTURE_KIT, FileExtensions, StatusOptions
from cg.constants.tb import AnalysisType
from cg.io.yaml import write_yaml
from cg.store.models import BedVersion, Case, CaseSample, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class MIPDNAConfigFileCreator:
    def __init__(self, lims_api: LimsAPI, root: str, store: Store):
        self.lims_api = lims_api
        self.root = root
        self.store = store

    def create(self, case_id: str, bed_flag: str | None, file_path: Path) -> None:
        provided_bed_file: str | None = self._get_bed_file_name(bed_flag) if bed_flag else None
        content: dict = self._get_content(provided_bed_file=provided_bed_file, case_id=case_id)
        write_yaml(
            content=content,
            file_path=file_path,
        )
        LOG.info(f"Created MIP-DNA config file for case {case_id} at {file_path}")

    def _get_content(self, provided_bed_file: str | None, case_id: str) -> dict:
        case: Case = self.store.get_case_by_internal_id_strict(case_id)
        samples_data = []
        for case_sample in case.links:
            sample_content: dict = self._get_sample_content(
                provided_bed_file=provided_bed_file, case_sample=case_sample
            )
            samples_data.append(sample_content)
        return {
            "case": case.internal_id,
            "default_gene_panels": case.panels,
            "samples": samples_data,
        }

    def _get_sample_content(self, provided_bed_file: str | None, case_sample: CaseSample) -> dict:
        case: Case = case_sample.case
        sample: Sample = case_sample.sample
        bed_file: str = provided_bed_file or self._get_sample_bed_file(case=case, sample=sample)
        mother: str = case_sample.mother.internal_id if case_sample.mother else "0"
        father: str = case_sample.father.internal_id if case_sample.father else "0"

        phenotype = case_sample.status
        if len(case.links) == 1 and case_sample.status == StatusOptions.UNKNOWN:
            phenotype = StatusOptions.UNAFFECTED.value

        return {
            "analysis_type": sample.prep_category,
            "capture_kit": bed_file,
            "expected_coverage": sample.application_version.application.min_sequencing_depth,
            "father": father,
            "mother": mother,
            "phenotype": phenotype,
            "sample_display_name": sample.name,
            "sample_id": sample.internal_id,
            "sex": sample.sex,
        }

    def _get_bed_file_name(self, bed_flag: str) -> str:
        if bed_flag.endswith(FileExtensions.BED):
            return bed_flag
        bed_version: BedVersion = self.store.get_bed_version_by_short_name_strict(bed_flag)
        return bed_version.filename

    def _get_sample_bed_file(self, case: Case, sample: Sample) -> str:
        if sample.prep_category == AnalysisType.WGS:
            return DEFAULT_CAPTURE_KIT
        return self._get_target_bed_from_lims(case)

    def _get_target_bed_from_lims(self, case: Case) -> str:
        """Get target bed filename from LIMS."""
        # Should the input to the function be the sample instead of the case?
        sample: Sample = case.samples[0]
        bed_shortname: str = self.lims_api.get_capture_kit_strict(
            sample.from_sample or sample.internal_id
        )
        bed_version: BedVersion = self.store.get_bed_version_by_short_name_strict(bed_shortname)
        return bed_version.filename



from cg.server.endpoints.sequencing_metrics.dtos import PacbioSequencingMetricsRequest
from cg.services.sample_run_metrics_service.dtos import (
    IlluminaSequencingMetricsDTO,
    PacbioSequencingMetricsDTO,
)
from cg.services.sample_run_metrics_service.utils import create_metrics_dto
from cg.store.models import IlluminaSequencingRun, PacbioSampleSequencingMetrics
from cg.store.store import Store


class SampleRunMetricsService:
    def __init__(self, store: Store):
        self.store = store

    def get_illumina_metrics(self, flow_cell_name: str) -> list[IlluminaSequencingMetricsDTO]:
        run: IlluminaSequencingRun = self.store.get_illumina_sequencing_run_by_device_internal_id(
            flow_cell_name
        )
        return create_metrics_dto(run.sample_metrics) if run else []

    def get_pacbio_metrics(
        self, metrics_request: PacbioSequencingMetricsRequest
    ) -> list[PacbioSequencingMetricsDTO]:
        response: list[PacbioSequencingMetricsDTO] = []
        sequencing_metrics: list[PacbioSampleSequencingMetrics] = (
            self.store.get_pacbio_sample_sequencing_metrics(
                sample_id=metrics_request.sample_id, smrt_cell_ids=metrics_request.smrt_cell_ids
            )
        )
        for db_metrics in sequencing_metrics:
            metrics = PacbioSequencingMetricsDTO(
                hifi_mean_read_length=db_metrics.hifi_mean_read_length,
                hifi_median_read_quality=db_metrics.hifi_median_read_quality,
                hifi_reads=db_metrics.hifi_reads,
                hifi_yield=db_metrics.hifi_yield,
                sample_id=db_metrics.sample.internal_id,
                smrt_cell_id=db_metrics.instrument_run.device.internal_id,
            )
            response.append(metrics)
        return response



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.fastq.constants import FastqDeliveryType
from cg.services.orders.validation.order_types.fastq.models.sample import FastqSample


class FastqOrder(OrderWithSamples):
    delivery_type: FastqDeliveryType
    samples: list[FastqSample]

    @property
    def enumerated_samples(self) -> enumerate[FastqSample]:
        return enumerate(self.samples)



from cg.server.dto.cases.requests import CasesRequest
from cg.store.models import Case, Customer
from cg.store.store import Store


class CaseWebService:

    def __init__(self, store: Store):
        self.store = store

    def get_cases(
        self, request: CasesRequest, customers: list[Customer] | None
    ) -> tuple[list[dict], int]:
        """Return cases with links for a customer from the database."""
        cases, total = self._get_cases(request=request, customers=customers)
        cases_with_links: list[dict] = [case.to_dict(links=True) for case in cases]
        return cases_with_links, total

    def _get_cases(
        self, request: CasesRequest, customers: list[Customer] | None
    ) -> tuple[list[Case], int]:
        """Get cases based on the provided filters."""
        return self.store.get_cases_by_customers_action_and_case_search(
            action=request.action,
            case_search=request.enquiry,
            customers=customers,
            offset=(request.page - 1) * request.page_size,
            limit=request.page_size,
        )



from cg.services.orders.validation.rules.sample.rules import (
    validate_application_compatibility,
    validate_application_exists,
    validate_applications_not_archived,
    validate_container_name_required,
    validate_sample_names_available,
    validate_sample_names_unique,
    validate_tube_container_name_unique,
    validate_volume_required,
    validate_well_position_format,
    validate_well_positions_required,
    validate_wells_contain_at_most_one_sample,
)

PACBIO_LONG_READ_SAMPLE_RULES: list[callable] = [
    validate_application_compatibility,
    validate_application_exists,
    validate_applications_not_archived,
    validate_container_name_required,
    validate_sample_names_available,
    validate_sample_names_unique,
    validate_tube_container_name_unique,
    validate_volume_required,
    validate_wells_contain_at_most_one_sample,
    validate_well_position_format,
    validate_well_positions_required,
]



"""Handler to update data objects."""

from datetime import datetime

from cg.constants import SequencingRunDataAvailability
from cg.constants.constants import CaseActions, SequencingQCStatus
from cg.constants.sequencing import Sequencers
from cg.services.illumina.post_processing.utils import get_q30_threshold
from cg.store.crud.read import ReadHandler
from cg.store.models import (
    Analysis,
    Case,
    IlluminaSampleSequencingMetrics,
    IlluminaSequencingRun,
    Order,
    PacbioSequencingRun,
    Sample,
)


class UpdateMixin(ReadHandler):
    """Contains methods to update database objects."""

    def update_sample_comment(self, sample_id: int, comment: str) -> None:
        """Update comment on sample with the provided comment."""
        sample: Sample = self.get_sample_by_entry_id(sample_id)
        sample.comment = f"{sample.comment} {comment}" if sample.comment else comment
        self.commit_to_store()

    def update_order_status(self, order_id: int, open: bool) -> Order:
        """Update the status of an order."""
        order: Order = self.get_order_by_id(order_id)
        order.is_open = open
        self.commit_to_store()
        return order

    def update_illumina_sequencing_run_data_availability(
        self,
        sequencing_run: IlluminaSequencingRun,
        data_availability: SequencingRunDataAvailability,
    ) -> IlluminaSequencingRun:
        """Update the data availability status of an Illumina sequencing run."""
        sequencing_run.data_availability = data_availability
        self.commit_to_store()
        return sequencing_run

    def update_illumina_sequencing_run_has_backup(
        self, sequencing_run: IlluminaSequencingRun, has_backup: bool
    ) -> IlluminaSequencingRun:
        """Update the backup status of an Illumina sequencing run."""
        sequencing_run.has_backup = has_backup
        self.commit_to_store()
        return sequencing_run

    def update_sequencing_qc_status(self, case: Case, status: SequencingQCStatus) -> None:
        case.aggregated_sequencing_qc = status
        self.commit_to_store()

    def update_case_action(self, action: CaseActions | None, case_internal_id: str) -> None:
        """Sets the action of the provided case the given action (can be None)."""
        case: Case = self.get_case_by_internal_id(internal_id=case_internal_id)
        case.action = action
        self.commit_to_store()

    def update_sample_reads_illumina(self, internal_id: str, sequencer_type: Sequencers):
        sample: Sample = self.get_sample_by_internal_id(internal_id)
        total_reads_for_sample: int = 0
        sample_metrics: list[IlluminaSampleSequencingMetrics] = sample.sample_run_metrics

        q30_threshold: int = get_q30_threshold(sequencer_type)

        for sample_metric in sample_metrics:
            if (
                sample_metric.base_passing_q30_percent >= q30_threshold
                or sample.is_negative_control
            ):
                total_reads_for_sample += sample_metric.total_reads_in_lane

        sample.reads = total_reads_for_sample
        self.commit_to_store()

    def update_sample_reads_pacbio(self, internal_id: str, reads: int):
        """Add reads to the current reads for a sample."""
        sample: Sample = self.get_sample_by_internal_id_strict(internal_id)
        sample.reads = reads
        self.commit_to_store()

    def update_sample_sequenced_at(self, internal_id: str, date: datetime):
        sample: Sample = self.get_sample_by_internal_id(internal_id)
        sample.last_sequenced_at = date
        self.commit_to_store()

    def mark_sample_as_cancelled(self, sample_id: int) -> None:
        sample: Sample = self.get_sample_by_entry_id(sample_id)
        sample.is_cancelled = True
        self.commit_to_store()

    def update_analysis_completed_at(self, analysis_id: int, completed_at: datetime | None) -> None:
        analysis: Analysis = self.get_analysis_by_entry_id(analysis_id)
        analysis.completed_at = completed_at
        self.commit_to_store()

    def update_analysis_housekeeper_version_id(self, analysis_id: int, version_id: int) -> None:
        """Update the Housekeeper version ID of an analysis."""
        analysis: Analysis = self.get_analysis_by_entry_id(analysis_id)
        analysis.housekeeper_version_id = version_id
        self.commit_to_store()

    def update_analysis_comment(self, analysis_id: int, comment: str | None) -> None:
        """Update the comment field of an analysis appending the new comment over any existing."""
        analysis: Analysis = self.get_analysis_by_entry_id(analysis_id)
        analysis.comment = f"{analysis.comment}\n{comment}" if analysis.comment else comment
        self.commit_to_store()

    def update_analysis_uploaded_at(self, analysis_id: int, uploaded_at: datetime | None) -> None:
        """Update the uploaded at field of an analysis."""
        analysis: Analysis = self.get_analysis_by_entry_id(analysis_id)
        analysis.uploaded_at = uploaded_at
        self.commit_to_store()

    def update_analysis_upload_started_at(
        self, analysis_id: int, upload_started_at: datetime | None
    ) -> None:
        """Update the upload started at field of an analysis."""
        analysis: Analysis = self.get_analysis_by_entry_id(analysis_id)
        analysis.upload_started_at = upload_started_at
        self.commit_to_store()

    def update_analysis_delivery_report_date(
        self, analysis_id: int, delivery_report_date: datetime | None
    ) -> None:
        """Update the delivery report created_at field of an analysis."""
        analysis: Analysis = self.get_analysis_by_entry_id(analysis_id)
        analysis.delivery_report_created_at = delivery_report_date
        self.commit_to_store()

    def update_pacbio_sequencing_run_comment(self, id: int, comment: str):
        sequencing_run: PacbioSequencingRun = self.get_pacbio_sequencing_run_by_id(id)
        sequencing_run.comment = comment
        self.commit_to_store()

    def update_pacbio_sequencing_run_processed(self, id: int, processed: bool):
        sequencing_run: PacbioSequencingRun = self.get_pacbio_sequencing_run_by_id(id)
        sequencing_run.processed = processed
        self.commit_to_store()



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_case import ExistingCase
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.models.sample import Sample

NewCaseType = Annotated[Case, Tag("new")]
ExistingCaseType = Annotated[ExistingCase, Tag("existing")]


class OrderWithCases(Order):
    cases: list[Annotated[NewCaseType | ExistingCaseType, Discriminator(has_internal_id)]]

    @property
    def enumerated_cases(self) -> enumerate[Case | ExistingCase]:
        return enumerate(self.cases)

    @property
    def enumerated_new_cases(self) -> list[tuple[int, Case]]:
        cases: list[tuple[int, Case]] = []
        for case_index, case in self.enumerated_cases:
            if case.is_new:
                cases.append((case_index, case))
        return cases

    @property
    def enumerated_existing_cases(self) -> list[tuple[int, ExistingCase]]:
        cases: list[tuple[int, ExistingCase]] = []
        for case_index, case in self.enumerated_cases:
            if not case.is_new:
                cases.append((case_index, case))
        return cases

    @property
    def enumerated_new_samples(self) -> list[tuple[int, int, Sample]]:
        return [
            (case_index, sample_index, sample)
            for case_index, case in self.enumerated_new_cases
            for sample_index, sample in case.enumerated_new_samples
        ]

    @property
    def enumerated_existing_samples(self) -> list[tuple[int, int, ExistingSample]]:
        return [
            (case_index, sample_index, sample)
            for case_index, case in self.enumerated_new_cases
            for sample_index, sample in case.enumerated_existing_samples
        ]



from datetime import datetime
from enum import Enum
from typing import Callable

from sqlalchemy import and_, not_, or_
from sqlalchemy.orm import Query

from cg.constants import REPORT_SUPPORTED_DATA_DELIVERY
from cg.constants.constants import CaseActions, DataDelivery, SequencingQCStatus, Workflow
from cg.constants.observations import (
    LOQUSDB_CANCER_SEQUENCING_METHODS,
    LOQUSDB_RARE_DISEASE_SEQUENCING_METHODS,
    LOQUSDB_SUPPORTED_WORKFLOWS,
)
from cg.store.models import Analysis, Application, Case, Customer, Sample


def filter_cases_by_action(cases: Query, action: str, **kwargs) -> Query:
    """Filter cases with matching action."""
    return cases.filter(Case.action == action) if action else cases


def filter_cases_by_case_search(cases: Query, case_search: str, **kwargs) -> Query:
    """Filter cases with matching internal id or name."""
    return (
        cases.filter(
            or_(
                Case.internal_id.contains(case_search),
                Case.name.contains(case_search),
            )
        )
        if case_search
        else cases
    )


def filter_cases_by_customer_entry_id(cases: Query, customer_entry_id: int, **kwargs) -> Query:
    """Filter cases with matching customer id."""
    return cases.filter(Case.customer_id == customer_entry_id)


def filter_cases_by_customer_entry_ids(
    cases: Query, customer_entry_ids: list[int], **kwargs
) -> Query:
    """Filter cases with matching customer ids."""
    return cases.filter(Case.customer_id.in_(customer_entry_ids)) if customer_entry_ids else cases


def filter_cases_by_entry_id(cases: Query, entry_id: int, **kwargs) -> Query:
    """Filter cases by entry id."""
    return cases.filter(Case.id == entry_id)


def filter_case_by_internal_id(cases: Query, internal_id: str, **kwargs) -> Query:
    """Filter cases with matching internal id."""
    return cases.filter(Case.internal_id == internal_id)


def filter_cases_by_internal_ids(cases: Query, internal_ids: list[str], **kwargs) -> Query:
    """Filter cases with internal ids."""
    return cases.filter(Case.internal_id.in_(internal_ids))


def filter_cases_by_internal_id_search(cases: Query, internal_id_search: str, **kwargs) -> Query:
    """Filter cases with internal ids matching the search pattern."""
    return cases.filter(Case.internal_id.contains(internal_id_search))


def filter_cases_by_name(cases: Query, name: str, **kwargs) -> Query:
    """Filter cases with matching name."""
    return cases.filter(Case.name == name) if name else cases


def filter_cases_by_name_search(cases: Query, name_search: str, **kwargs) -> Query:
    """Filter cases with names matching the search pattern."""
    return cases.filter(Case.name.contains(name_search))


def filter_cases_by_workflows(cases: Query, workflows: list[Workflow], **kwargs) -> Query:
    """Filter cases by data analysis types."""
    return cases.filter(Case.data_analysis.in_(workflows))


def filter_cases_by_workflow_search(cases: Query, workflow_search: str, **kwargs) -> Query:
    """Filter cases with a workflow search pattern."""
    return cases.filter(Case.data_analysis.contains(workflow_search))


def filter_cases_by_priority(cases: Query, priority: str, **kwargs) -> Query:
    """Filter cases with matching priority."""
    return cases.filter(Case.priority == priority)


def filter_cases_for_analysis(cases: Query, **kwargs) -> Query:
    """Filter cases that should be analyzed:
    1. All cases with action set to ANALYZE should be analyzed, regardless of other conditions.
    2. A case that has not been analyzed before
    3. A microSALT case with at least one sample has new data that has not been analyzed
    4. A case that has been set to TOP_UP by production and has new data
    """
    return cases.filter(
        or_(
            # Overriding state: Analyse no matter what
            Case.action == CaseActions.ANALYZE,
            # Case has not been analysed before
            and_(
                Application.is_external.isnot(True),
                Case.action.is_(None),
                Analysis.created_at.is_(None),
            ),
            # Case contains new data that has not been analysed. (Only relevant for microSALT)
            and_(
                Case.action.is_(None),
                Analysis.created_at < Sample.last_sequenced_at,
                Case.data_analysis == Workflow.MICROSALT,
            ),
            # Cases manually set to top-up by production that get the new data
            and_(
                Case.action == CaseActions.TOP_UP,
                Analysis.created_at < Sample.last_sequenced_at,
            ),
        )
    )


def filter_cases_has_sequence(cases: Query, **kwargs) -> Query:
    """Filter cases that are not sequenced according to record in StatusDB."""
    return cases.filter(or_(Application.is_external, Sample.last_sequenced_at.isnot(None)))


def filter_cases_not_analysed(cases: Query, **kwargs) -> Query:
    """Filter cases that have not been analysed and are not currently being analysed."""
    not_analyzed_condition = not_(Case.analyses.any(Analysis.completed_at.isnot(None)))
    not_in_progress_condition = Case.action != CaseActions.ANALYZE

    return cases.filter(and_(not_analyzed_condition, not_in_progress_condition))


def filter_cases_with_workflow(cases: Query, workflow: Workflow = None, **kwargs) -> Query:
    """Filter cases with workflow."""
    return cases.filter(Case.data_analysis == workflow) if workflow else cases


def filter_cases_with_loqusdb_supported_workflow(
    cases: Query, workflow: Workflow = None, **kwargs
) -> Query:
    """Filter Loqusdb related cases with workflow."""
    records: Query = (
        cases.filter(Case.data_analysis == workflow)
        if workflow
        else cases.filter(Case.data_analysis.in_(LOQUSDB_SUPPORTED_WORKFLOWS))
    )
    return records.filter(Customer.loqus_upload)


def filter_cases_with_loqusdb_supported_sequencing_method(
    cases: Query, workflow: Workflow = None, **kwargs
) -> Query:
    """Filter cases with Loqusdb supported sequencing method."""
    supported_sequencing_methods = {
        Workflow.MIP_DNA: LOQUSDB_RARE_DISEASE_SEQUENCING_METHODS,
        Workflow.BALSAMIC: LOQUSDB_CANCER_SEQUENCING_METHODS,
    }
    return (
        cases.filter(Application.prep_category.in_(supported_sequencing_methods[workflow]))
        if workflow
        else cases
    )


def filter_cases_with_scout_data_delivery(cases: Query, **kwargs) -> Query:
    """Filter cases containing Scout as a data delivery option."""
    return cases.filter(Case.data_delivery.contains(DataDelivery.SCOUT))


def filter_newer_cases_by_order_date(cases: Query, order_date: datetime, **kwargs) -> Query:
    """Filter cases newer than date."""
    cases: Query = cases.filter(Case.ordered_at > order_date)
    return cases.order_by(Case.ordered_at.asc())


def filter_inactive_analysis_cases(cases: Query, **kwargs) -> Query:
    """Filter cases which are not set or on hold."""
    return cases.filter(
        or_(
            Case.action.is_(None),
            Case.action == CaseActions.HOLD,
        )
    )


def filter_older_cases_by_creation_date(cases: Query, creation_date: datetime, **kwargs) -> Query:
    """Filter older cases compared to date."""
    cases = cases.filter(Case.created_at < creation_date)
    return cases.order_by(Case.created_at.asc())


def filter_report_supported_data_delivery_cases(cases: Query, **kwargs) -> Query:
    """Filter cases with a valid data delivery for delivery report generation."""
    return cases.filter(Case.data_delivery.in_(REPORT_SUPPORTED_DATA_DELIVERY))


def filter_running_cases(cases: Query, **kwargs) -> Query:
    """Filter cases which are running."""
    return cases.filter(Case.action == CaseActions.RUNNING)


def filter_compressible_cases(cases: Query, **kwargs) -> Query:
    """Filter cases which are running."""
    return cases.filter(Case.is_compressible)


def order_cases_by_created_at(cases: Query, **kwargs) -> Query:
    """Order cases by created at."""
    return cases.order_by(Case.created_at.desc())


def filter_cases_pending_or_failed_sequencing_qc(cases: Query, **kwargs) -> Query:
    """Filter cases with pending or failed sequencing QC."""
    return cases.filter(
        or_(
            Case.aggregated_sequencing_qc == SequencingQCStatus.PENDING,
            Case.aggregated_sequencing_qc == SequencingQCStatus.FAILED,
        )
    )


def filter_cases_passing_sequencing_qc(cases: Query, **kwargs) -> Query:
    """Filter cases with passing sequencing QC."""
    return cases.filter(Case.aggregated_sequencing_qc == SequencingQCStatus.PASSED)


def apply_case_filter(
    cases: Query,
    filter_functions: list[Callable],
    action: str | None = None,
    case_search: str | None = None,
    creation_date: datetime | None = None,
    customer_entry_id: int | None = None,
    customer_entry_ids: list[int] | None = None,
    entry_id: int | None = None,
    internal_id: str | None = None,
    internal_ids: list[str] | None = None,
    internal_id_search: str | None = None,
    name: str | None = None,
    name_search: str | None = None,
    order_date: datetime | None = None,
    workflow: Workflow | None = None,
    workflow_search: str | None = None,
    workflows: list[Workflow] | None = None,
    priority: str | None = None,
    ticket_id: str | None = None,
    order_id: int | None = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        cases: Query = function(
            action=action,
            case_search=case_search,
            cases=cases,
            creation_date=creation_date,
            customer_entry_id=customer_entry_id,
            customer_entry_ids=customer_entry_ids,
            entry_id=entry_id,
            internal_id=internal_id,
            internal_ids=internal_ids,
            internal_id_search=internal_id_search,
            name=name,
            name_search=name_search,
            order_date=order_date,
            workflow=workflow,
            workflow_search=workflow_search,
            workflows=workflows,
            priority=priority,
            ticket_id=ticket_id,
            order_id=order_id,
        )
    return cases


class CaseFilter(Enum):
    """Define case filters."""

    BY_ACTION: Callable = filter_cases_by_action
    BY_CASE_SEARCH: Callable = filter_cases_by_case_search
    BY_CUSTOMER_ENTRY_ID: Callable = filter_cases_by_customer_entry_id
    BY_CUSTOMER_ENTRY_IDS: Callable = filter_cases_by_customer_entry_ids
    BY_ENTRY_ID: Callable = filter_cases_by_entry_id
    BY_INTERNAL_ID: Callable = filter_case_by_internal_id
    BY_INTERNAL_IDS: Callable = filter_cases_by_internal_ids
    BY_INTERNAL_ID_SEARCH: Callable = filter_cases_by_internal_id_search
    BY_NAME: Callable = filter_cases_by_name
    BY_NAME_SEARCH: Callable = filter_cases_by_name_search
    BY_WORKFLOWS: Callable = filter_cases_by_workflows
    BY_WORKFLOW_SEARCH: Callable = filter_cases_by_workflow_search
    BY_PRIORITY: Callable = filter_cases_by_priority
    FOR_ANALYSIS: Callable = filter_cases_for_analysis
    HAS_INACTIVE_ANALYSIS: Callable = filter_inactive_analysis_cases
    HAS_SEQUENCE: Callable = filter_cases_has_sequence
    IS_RUNNING: Callable = filter_running_cases
    IS_COMPRESSIBLE: Callable = filter_compressible_cases
    NEW_BY_ORDER_DATE: Callable = filter_newer_cases_by_order_date
    NOT_ANALYSED: Callable = filter_cases_not_analysed
    OLD_BY_CREATION_DATE: Callable = filter_older_cases_by_creation_date
    REPORT_SUPPORTED: Callable = filter_report_supported_data_delivery_cases
    WITH_LOQUSDB_SUPPORTED_WORKFLOW: Callable = filter_cases_with_loqusdb_supported_workflow
    WITH_LOQUSDB_SUPPORTED_SEQUENCING_METHOD: Callable = (
        filter_cases_with_loqusdb_supported_sequencing_method
    )
    WITH_WORKFLOW: Callable = filter_cases_with_workflow
    WITH_SCOUT_DELIVERY: Callable = filter_cases_with_scout_data_delivery
    ORDER_BY_CREATED_AT: Callable = order_cases_by_created_at
    PENDING_OR_FAILED_SEQUENCING_QC: Callable = filter_cases_pending_or_failed_sequencing_qc
    PASSING_SEQUENCING_QC: Callable = filter_cases_passing_sequencing_qc



from cg.constants.devices import DeviceType
from cg.services.run_devices.pacbio.data_transfer_service.dto import (
    PacBioSampleSequencingMetricsDTO,
    PacBioSequencingRunDTO,
    PacBioSMRTCellDTO,
    PacBioSMRTCellMetricsDTO,
)
from cg.services.run_devices.pacbio.metrics_parser.models import PacBioMetrics, SampleMetrics
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData


def get_smrt_cell_dto(metrics: PacBioMetrics) -> PacBioSMRTCellDTO:
    internal_id: str = metrics.dataset_metrics.cell_id
    return PacBioSMRTCellDTO(type=DeviceType.PACBIO, internal_id=internal_id)


def get_smrt_cell_metrics_dto(
    metrics: PacBioMetrics, run_data: PacBioRunData
) -> PacBioSMRTCellMetricsDTO:
    hifi_mean_read_quality: str = f"Q{metrics.read.hifi_median_read_quality}"
    return PacBioSMRTCellMetricsDTO(
        type=DeviceType.PACBIO,
        well=metrics.dataset_metrics.well,
        plate=metrics.dataset_metrics.plate,
        run_id=run_data.run_id,
        started_at=metrics.dataset_metrics.run_started_at,
        completed_at=metrics.dataset_metrics.run_completed_at,
        hifi_reads=metrics.read.hifi_reads,
        hifi_yield=metrics.read.hifi_yield,
        hifi_mean_read_length=metrics.read.hifi_mean_read_length,
        hifi_median_read_length=metrics.read.hifi_median_read_length,
        hifi_median_read_quality=hifi_mean_read_quality,
        hifi_mean_length_n50=metrics.read.hifi_mean_length_n50,
        percent_reads_passing_q30=metrics.read.percent_q30,
        productive_zmws=metrics.productivity.productive_zmws,
        p0_percent=metrics.productivity.percent_p_0,
        p1_percent=metrics.productivity.percent_p_1,
        p2_percent=metrics.productivity.percent_p_2,
        polymerase_mean_read_length=metrics.polymerase.mean_read_length,
        polymerase_read_length_n50=metrics.polymerase.read_length_n50,
        polymerase_mean_longest_subread=metrics.polymerase.mean_longest_subread_length,
        polymerase_longest_subread_n50=metrics.polymerase.longest_subread_length_n50,
        control_reads=metrics.control.reads,
        control_mean_read_length=metrics.control.mean_read_length,
        control_mean_read_concordance=metrics.control.percent_mean_concordance_reads,
        control_mode_read_concordance=metrics.control.percent_mode_concordance_reads,
        failed_reads=metrics.read.failed_reads,
        failed_yield=metrics.read.failed_yield,
        failed_mean_read_length=metrics.read.failed_mean_read_length,
        barcoded_hifi_reads=metrics.barcodes.barcoded_hifi_reads,
        barcoded_hifi_reads_percentage=metrics.barcodes.barcoded_hifi_reads_percentage,
        barcoded_hifi_yield=metrics.barcodes.barcoded_hifi_yield,
        barcoded_hifi_yield_percentage=metrics.barcodes.barcoded_hifi_yield_percentage,
        barcoded_hifi_mean_read_length=metrics.barcodes.barcoded_hifi_mean_read_length,
        unbarcoded_hifi_reads=metrics.barcodes.unbarcoded_hifi_reads,
        unbarcoded_hifi_yield=metrics.barcodes.unbarcoded_hifi_yield,
        unbarcoded_hifi_mean_read_length=metrics.barcodes.unbarcoded_hifi_mean_read_length,
        movie_name=metrics.dataset_metrics.movie_name,
    )


def get_sample_sequencing_metrics_dtos(
    sample_metrics: list[SampleMetrics],
) -> list[PacBioSampleSequencingMetricsDTO]:
    sample_metrics_dtos: list[PacBioSampleSequencingMetricsDTO] = []
    for sample in sample_metrics:
        sample_sequencing_metrics_dto = PacBioSampleSequencingMetricsDTO(
            sample_internal_id=sample.sample_internal_id,
            hifi_reads=sample.hifi_reads,
            hifi_yield=sample.hifi_yield,
            hifi_mean_read_length=sample.hifi_mean_read_length,
            hifi_median_read_quality=sample.hifi_median_read_quality,
            polymerase_mean_read_length=sample.polymerase_read_length,
        )
        sample_metrics_dtos.append(sample_sequencing_metrics_dto)
    return sample_metrics_dtos


def get_sequencing_run_dto(
    metrics: PacBioMetrics, run_data: PacBioRunData
) -> PacBioSequencingRunDTO:
    return PacBioSequencingRunDTO(
        instrument_name=metrics.dataset_metrics.instrument_name,  # type: ignore - pydantic will convert to correct type
        run_id=run_data.run_id,
    )



import logging
import tempfile
from datetime import date, datetime, timedelta
from json.decoder import JSONDecodeError
from pathlib import Path

from cg.apps.crunchy.models import CrunchyFile, CrunchyMetadata
from cg.constants import FASTQ_DELTA
from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile, ReadStream, WriteFile
from cg.utils.date import get_date

LOG = logging.getLogger(__name__)

NR_OF_FILES_IN_METADATA = 3


# Methods to get file information
def get_log_dir(file_path: Path) -> Path:
    """Return the path to where logs should be stored"""
    return file_path.parent


def get_fastq_to_spring_sbatch_path(log_dir: Path, run_name: str = None) -> Path:
    """Return the path to where compression sbatch should be printed"""
    return Path(log_dir, "_".join([run_name, "compress_fastq.sh"]))


def get_spring_to_fastq_sbatch_path(log_dir: Path, run_name: str = None) -> Path:
    """Return the path to where decompression sbatch should be printed"""
    return Path(log_dir, "_".join([run_name, "decompress_spring.sh"]))


def get_tmp_dir(prefix: str, suffix: str, base: str = None) -> str:
    """Create a temporary directory and return the path to it"""

    with tempfile.TemporaryDirectory(prefix=prefix, suffix=suffix, dir=base) as dir_name:
        tmp_dir_path = dir_name

    LOG.info(f"Created temporary dir {tmp_dir_path}")
    return tmp_dir_path


def get_crunchy_metadata(metadata_path: Path) -> CrunchyMetadata:
    """Validate content of metadata file and return mapped content."""
    LOG.info(f"Fetch SPRING metadata from {metadata_path}")
    try:
        content: list[dict[str, str]] = ReadFile.get_content_from_file(
            file_format=FileFormat.JSON, file_path=metadata_path
        )
    except JSONDecodeError:
        message = "No content in SPRING metadata file"
        LOG.warning(message)
        raise SyntaxError(message)
    metadata = CrunchyMetadata(files=content)

    if len(metadata.files) != NR_OF_FILES_IN_METADATA:
        LOG.warning(f"Wrong number of files in SPRING metadata file: {metadata_path}")
        LOG.info(
            f"Found {len(metadata.files)} files, should always be {NR_OF_FILES_IN_METADATA} files"
        )
        raise SyntaxError

    return metadata


def get_file_updated_at(crunchy_metadata: CrunchyMetadata) -> date | None:
    """Check if a SPRING metadata file has been updated and return the date when updated"""
    return crunchy_metadata.files[0].updated


def get_spring_archive_files(crunchy_metadata: CrunchyMetadata) -> dict[str, CrunchyFile]:
    """Map the files in SPRING metadata to a dictionary

    Returns: {
                "fastq_first" : {file_info},
                "fastq_second" : {file_info},
                "spring" : {file_info},
              }
    """
    names_map = {"first_read": "fastq_first", "second_read": "fastq_second", "spring": "spring"}
    archive_files = {}
    for file_info in crunchy_metadata.files:
        file_name = names_map[file_info.file]
        archive_files[file_name] = file_info
    return archive_files


def write_metadata_content(spring_metadata: CrunchyMetadata, spring_metadata_path: Path) -> None:
    """Update the file on disk."""
    content: dict = ReadStream.get_content_from_stream(
        file_format=FileFormat.JSON, stream=spring_metadata.json(exclude_none=True)
    )
    WriteFile.write_file_from_content(
        content=content["files"], file_format=FileFormat.JSON, file_path=spring_metadata_path
    )


def update_metadata_date(spring_metadata_path: Path) -> None:
    """Update date in the SPRING metadata file to today date"""
    now: datetime = get_date()
    spring_metadata: CrunchyMetadata = get_crunchy_metadata(spring_metadata_path)
    LOG.info("Adding today date to SPRING metadata file")
    for file_info in spring_metadata.files:
        file_info.updated = now.date()
    write_metadata_content(
        spring_metadata=spring_metadata, spring_metadata_path=spring_metadata_path
    )


def update_metadata_paths(spring_metadata_path: Path, new_parent_path: Path) -> None:
    """Update paths for file in SPRING metadata file."""
    spring_metadata: CrunchyMetadata = get_crunchy_metadata(metadata_path=spring_metadata_path)
    LOG.info(f"Updating file paths in SPRING metadata file: {spring_metadata_path}")
    for file in spring_metadata.files:
        file.path = Path(new_parent_path, Path(file.path).name)
    write_metadata_content(
        spring_metadata=spring_metadata, spring_metadata_path=spring_metadata_path
    )


def check_if_update_spring(file_date: date) -> bool:
    """Check if date is older than FASTQ_DELTA."""
    delta = file_date + timedelta(days=FASTQ_DELTA)
    now = datetime.now()
    if delta > now.date():
        LOG.info("FASTQ files are not old enough")
        return False
    return True



"""Utils for sample sheet generation."""

import logging
from pathlib import Path
from housekeeper.store.models import File
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.utils.files import get_file_in_directory

LOG = logging.getLogger(__name__)


def add_and_include_sample_sheet_path_to_housekeeper(
    flow_cell_directory: Path, flow_cell_name: str, hk_api: HousekeeperAPI
) -> None:
    """
    Add sample sheet path to Housekeeper.
    Raises:
        FileNotFoundError: If the sample sheet file is not found.
    """
    LOG.debug("Adding sample sheet to Housekeeper")
    try:
        sample_sheet_file_path: Path = get_sample_sheet_path_from_flow_cell_dir(flow_cell_directory)
        hk_api.add_bundle_and_version_if_non_existent(flow_cell_name)
        hk_api.add_file_to_bundle_if_non_existent(
            file_path=sample_sheet_file_path,
            bundle_name=flow_cell_name,
            tag_names=[SequencingFileTag.SAMPLE_SHEET, flow_cell_name],
        )
    except FileNotFoundError as e:
        LOG.error(
            f"Sample sheet for flow cell {flow_cell_name} in {flow_cell_directory} was not found, error: {e}"
        )


def delete_sample_sheet_from_housekeeper(flow_cell_id: str, hk_api: HousekeeperAPI) -> None:
    """Delete a sample sheet from Housekeeper database and disk given its path."""
    sample_sheet_file_path: Path = hk_api.get_sample_sheet_path(flow_cell_id)
    file: File = hk_api.get_file_insensitive_path(sample_sheet_file_path)
    hk_api.delete_file(file_id=file.id)


def get_sample_sheet_path_from_flow_cell_dir(
    flow_cell_directory: Path,
    sample_sheet_file_name: str = DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME,
) -> Path:
    """Return the path to the sample sheet in the flow cell directory."""
    return get_file_in_directory(directory=flow_cell_directory, file_name=sample_sheet_file_name)



from pathlib import Path

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.exc import AnalysisNotReadyError
from cg.services.analysis_starter.input_fetcher.input_fetcher import InputFetcher
from cg.store.models import Case
from cg.store.store import Store


class BamFetcher(InputFetcher):
    def __init__(self, housekeeper_api: HousekeeperAPI, status_db: Store) -> None:
        self.housekeeper_api = housekeeper_api
        self.status_db = status_db

    def ensure_files_are_ready(self, case_id: str) -> None:
        case: Case = self.status_db.get_case_by_internal_id_strict(case_id)
        samples_without_files: list[str] = self._get_samples_without_files(case)
        missing_files: list[str] = self._get_missing_file_paths(case)
        if samples_without_files or missing_files:
            self._raise_error(
                missing_files=missing_files, samples_without_files=samples_without_files
            )

    def _get_samples_without_files(self, case: Case) -> list[str]:
        samples_without_files: list[str] = []
        for sample in case.samples:
            sample_files: list[File] = self.housekeeper_api.files(
                bundle=sample.internal_id, tags={"bam"}
            ).all()
            if not sample_files:
                samples_without_files.append(sample.internal_id)
        return samples_without_files

    def _get_missing_file_paths(self, case: Case) -> list[str]:
        missing_files: list[str] = []
        for sample in case.samples:
            sample_files: list[File] = self.housekeeper_api.files(
                bundle=sample.internal_id, tags={"bam"}
            ).all()
            missing_sample_files: list[str] = [
                file.full_path for file in sample_files if not Path(file.full_path).is_file()
            ]
            missing_files.extend(missing_sample_files)
        return missing_files

    @staticmethod
    def _raise_error(missing_files: list[str], samples_without_files: list[str]) -> None:
        error_message = ""
        if samples_without_files:
            samples_without_files_str = "\n".join(samples_without_files)
            error_message += f"The following samples are missing BAM files in Housekeeper: \n{samples_without_files_str}"
        if missing_files:
            missing_files_str = "\n".join(missing_files)
            error_message += f"\nThe following BAM files are missing on disk: \n{missing_files_str}"
        raise AnalysisNotReadyError(error_message)



"""Module for Taxprofiler Analysis API."""

import logging
from pathlib import Path

from cg.constants import Workflow
from cg.constants.constants import GenomeVersion
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase
from cg.models.taxprofiler.taxprofiler import TaxprofilerQCMetrics
from cg.resources import TAXPROFILER_BUNDLE_FILENAMES_PATH
from cg.store.models import Sample

LOG = logging.getLogger(__name__)


class TaxprofilerAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Taxprofiler processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.TAXPROFILER,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.taxprofiler.root
        self.workflow_bin_path: str = config.taxprofiler.workflow_bin_path
        self.profile: str = config.taxprofiler.profile
        self.conda_env: str = config.taxprofiler.conda_env
        self.conda_binary: str = config.taxprofiler.conda_binary
        self.platform: str = config.taxprofiler.platform
        self.params: str = config.taxprofiler.params
        self.workflow_config_path: str = config.taxprofiler.config
        self.resources: str = config.taxprofiler.resources
        self.revision: str = config.taxprofiler.revision
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.taxprofiler.tower_workflow
        self.account: str = config.taxprofiler.slurm.account
        self.email: str = config.taxprofiler.slurm.mail_user

    @property
    def is_multiqc_pattern_search_exact(self) -> bool:
        """Only exact pattern search is allowed to collect metrics information from multiqc file."""
        return True

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Taxprofiler bundle filenames path."""
        return TAXPROFILER_BUNDLE_FILENAMES_PATH

    def get_multiqc_search_patterns(self, case_id: str) -> dict:
        """Return search patterns for MultiQC for Taxprofiler."""
        samples: list[Sample] = self.status_db.get_samples_by_case_id(case_id=case_id)
        search_patterns: dict[str, str] = {
            f"{sample.name}_{sample.name}": sample.internal_id for sample in samples
        }
        return search_patterns

    def get_genome_build(self, case_id: str) -> str:
        """Return the reference genome build version of a Taxprofiler analysis."""
        return GenomeVersion.T2T_CHM13.value

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> NextflowAnalysis:
        """Parse Nextflow output analysis files and return an analysis model."""
        qc_metrics_model = TaxprofilerQCMetrics
        return super().parse_analysis(
            qc_metrics_raw=qc_metrics_raw, qc_metrics_model=qc_metrics_model, **kwargs
        )



"""Store class that combines all CRUD operations for the CG database."""

from cg.store.crud.create import CreateMixin
from cg.store.crud.delete import DeleteMixin
from cg.store.crud.update import UpdateMixin


class Store(
    CreateMixin,
    DeleteMixin,
    UpdateMixin,
):
    def recalculate_sample_reads_pacbio(self, sample_id: str) -> None:
        reads: int = sum(
            metric.hifi_reads
            for metric in self.get_pacbio_sample_sequencing_metrics(
                sample_id=sample_id, smrt_cell_ids=None
            )
        )
        self.update_sample_reads_pacbio(internal_id=sample_id, reads=reads)



"""
Code to handle communications to the shell from CG.
"""

import copy
import logging
import subprocess
from subprocess import CalledProcessError

from cg.constants.process import EXIT_SUCCESS

LOG = logging.getLogger(__name__)


class Process:
    """Class to handle communication with other programs via the shell.

    The other parts of the code should not need to have any knowledge about how the processes are
    called; that will be handled in this module.Output form stdout and stdin will be handled here.
    """

    def __init__(
        self,
        binary: str,
        conda_binary: str = None,
        config: str = None,
        config_parameter: str = "--config",
        environment: str = None,
        launch_directory: str = None,
    ):
        """
        Args:
            binary(str): Path to binary for the process to use
            config(str): Path to config if used by process
            environment(str): Activate conda environment before executing binary
        """
        super(Process, self).__init__()
        self.binary: str = binary
        self.conda_binary: str = conda_binary
        self.config: str = config
        self.environment: str = environment
        self.launch_directory: str = launch_directory
        LOG.debug("Initialising Process with binary: %s", self.binary)
        self.base_call: list[str] = [self.binary]

        if conda_binary:
            LOG.debug(f"Activating environment with conda run for binary: {self.conda_binary}")
            self.base_call.insert(0, f"{self.conda_binary} run --name {self.environment}")
        elif environment:
            LOG.debug(f"Activating environment with: {self.environment}")
            self.base_call.insert(0, f"source activate {self.environment};")
        if config:
            self.base_call.extend([config_parameter, config])
        if launch_directory:
            self.base_call.insert(0, f"cd {self.launch_directory};")

        LOG.debug(f"Use base call {self.base_call}")
        self._stdout = ""
        self._stderr = ""

    def run_command(self, parameters: list = None, dry_run: bool = False) -> int:
        """Execute a command in the shell.
        If environment is supplied - shell=True has to be supplied to enable passing as a string for executing multiple
         commands

        Args:
            parameters(list):
            dry_run(bool): Print command instead of executing it
        Return(int): Return code from called process

        """
        command = copy.deepcopy(self.base_call)
        if parameters:
            command.extend(parameters)

        LOG.info("Running command %s", " ".join(command))
        if dry_run:
            LOG.info("Dry run: process call will not be executed!!")
            return EXIT_SUCCESS

        if self.environment:
            res = subprocess.run(
                " ".join(command),
                shell=True,
                check=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        else:
            res = subprocess.run(
                command, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )

        self.stdout = res.stdout.decode("utf-8").rstrip()
        self.stderr = res.stderr.decode("utf-8").rstrip()
        if res.returncode != EXIT_SUCCESS:
            LOG.critical("Call %s exit with a non zero exit code", command)
            LOG.critical(self.stderr)
            raise CalledProcessError(res.returncode, command)

        return res.returncode

    @property
    def stdout(self):
        """Fetch stdout"""
        return self._stdout

    @stdout.setter
    def stdout(self, text):
        self._stdout = text

    @stdout.deleter
    def stdout(self):
        del self._stdout

    @property
    def stderr(self):
        """Fetch stderr"""
        return self._stderr

    @stderr.setter
    def stderr(self, text):
        self._stderr = text

    @stderr.deleter
    def stderr(self):
        del self._stderr

    def stdout_lines(self):
        """Iterate over the lines in self.stdout"""
        yield from self.stdout.split("\n")

    def stderr_lines(self):
        """Iterate over the lines in self.stderr"""
        yield from self.stderr.split("\n")

    def __repr__(self):
        return f"Process:base_call:{self.base_call}"



from pathlib import Path

from cg.services.run_devices.abstract_classes import RunDataGenerator
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import PostProcessingRunDataGeneratorError
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.services.run_devices.validators import validate_has_expected_parts, validate_name_pre_fix


class PacBioRunDataGenerator(RunDataGenerator):

    @staticmethod
    def _validate_run_name(run_name) -> None:
        validate_name_pre_fix(run_name, pre_fix="r")
        validate_has_expected_parts(run_name=run_name, expected_parts=2)

    @handle_post_processing_errors(
        to_except=(ValueError,), to_raise=PostProcessingRunDataGeneratorError
    )
    def get_run_data(self, run_full_name: str, sequencing_dir: str) -> PacBioRunData:
        """
        Get the run data for a PacBio SMRT cell run.
        run_full_name should include the PacBio run including plate well, e.g. 'r84202_20240522_133539/1_A01'
        """
        self._validate_run_name(run_full_name)
        full_path = Path(sequencing_dir, run_full_name)

        return PacBioRunData(
            full_path=full_path,
            run_id=self._get_run_id_from_run_full_name(run_full_name),
            well_name=self._get_well(run_full_name),
            plate=self._get_plate(run_full_name),  # type: ignore Pydantic transforming
        )

    @staticmethod
    def _get_run_id_from_run_full_name(
        run_full_name: str,
    ) -> str:
        return run_full_name.split("/")[0]

    @staticmethod
    def _get_plate_well(run_name: str) -> str:
        return run_name.split("/")[1]

    def _get_plate(self, run_name: str) -> str:
        plate_well: str = self._get_plate_well(run_name)
        return plate_well.split("_")[0]

    def _get_well(self, run_name: str) -> str:
        plate_well: str = self._get_plate_well(run_name)
        return plate_well.split("_")[-1]



from http import HTTPStatus

from flask import Blueprint, Response, jsonify, request
from pydantic import ValidationError

from cg.server.endpoints.error_handler import handle_missing_entries
from cg.server.endpoints.sequencing_run.dtos import (
    PacbioSequencingRunResponse,
    PacbioSequencingRunUpdateRequest,
)
from cg.server.endpoints.utils import before_request
from cg.server.ext import pacbio_sequencing_runs_service

PACBIO_SEQUENCING_RUNS_BLUEPRINT = Blueprint(
    "pacbio_sequencing_runs", __name__, url_prefix="/api/v1/"
)
PACBIO_SEQUENCING_RUNS_BLUEPRINT.before_request(before_request)


@PACBIO_SEQUENCING_RUNS_BLUEPRINT.route("/pacbio_sequencing_runs", methods=["GET"])
@handle_missing_entries
def get_sequencing_runs():
    page: int = int(request.args.get("page", "0"))
    page_size: int = int(request.args.get("pageSize", "0"))
    sequencing_runs: PacbioSequencingRunResponse = (
        pacbio_sequencing_runs_service.get_sequencing_runs(page=page, page_size=page_size)
    )
    return jsonify(sequencing_runs.model_dump())


@PACBIO_SEQUENCING_RUNS_BLUEPRINT.route("/pacbio_sequencing_runs/<id>", methods=["PATCH"])
@handle_missing_entries
def update_sequencing_run(id: str):
    try:
        update_request = PacbioSequencingRunUpdateRequest(id=id, **request.json)
    except ValidationError:
        return Response(status=HTTPStatus.UNPROCESSABLE_ENTITY)
    pacbio_sequencing_runs_service.update_sequencing_run(update_request=update_request)
    return Response(status=HTTPStatus.NO_CONTENT)



"""Module for the Pacbio data transfer object creation service."""

import logging

from pydantic import ValidationError

from cg.services.run_devices.abstract_classes import PostProcessingDataTransferService
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import (
    PostProcessingDataTransferError,
    PostProcessingRunFileManagerError,
)
from cg.services.run_devices.pacbio.data_transfer_service.dto import (
    PacBioDTOs,
    PacBioSampleSequencingMetricsDTO,
    PacBioSequencingRunDTO,
    PacBioSMRTCellDTO,
    PacBioSMRTCellMetricsDTO,
)
from cg.services.run_devices.pacbio.data_transfer_service.utils import (
    get_sample_sequencing_metrics_dtos,
    get_sequencing_run_dto,
    get_smrt_cell_dto,
    get_smrt_cell_metrics_dto,
)
from cg.services.run_devices.pacbio.metrics_parser.metrics_parser import PacBioMetricsParser
from cg.services.run_devices.pacbio.metrics_parser.models import PacBioMetrics
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData

LOG = logging.getLogger(__name__)


class PacBioDataTransferService(PostProcessingDataTransferService):
    def __init__(self, metrics_service: PacBioMetricsParser):
        self.metrics_service: PacBioMetricsParser = metrics_service

    @handle_post_processing_errors(
        to_except=(PostProcessingRunFileManagerError, ValidationError),
        to_raise=PostProcessingDataTransferError,
    )
    def get_post_processing_dtos(self, run_data: PacBioRunData) -> PacBioDTOs:
        metrics: PacBioMetrics = self.metrics_service.parse_metrics(run_data)
        sequencing_run_dto: PacBioSequencingRunDTO = get_sequencing_run_dto(
            metrics=metrics, run_data=run_data
        )
        smrt_cell_dto: PacBioSMRTCellDTO = get_smrt_cell_dto(metrics)
        smrt_cell_metrics_dto: PacBioSMRTCellMetricsDTO = get_smrt_cell_metrics_dto(
            metrics=metrics, run_data=run_data
        )
        sample_sequencing_metrics_dtos: list[PacBioSampleSequencingMetricsDTO] = (
            get_sample_sequencing_metrics_dtos(metrics.samples)
        )
        LOG.debug("DTOs created")
        return PacBioDTOs(
            run_device=smrt_cell_dto,
            smrt_cell_metrics=smrt_cell_metrics_dto,
            sample_sequencing_metrics=sample_sequencing_metrics_dtos,
            sequencing_run=sequencing_run_dto,
        )



""" Ask questions to the shell """

import getpass
import os


def environ_email():
    """Guess email from sudo user environment variable or login name."""
    username = os.environ.get("SUDO_USER") or getpass.getuser()

    return f"{username}@scilifelab.se"



"""Module for Raredisease Analysis API."""

import logging
from itertools import permutations
from pathlib import Path
from typing import Any

from housekeeper.store.models import File

from cg.clients.chanjo2.models import (
    CoverageMetrics,
    CoveragePostRequest,
    CoveragePostResponse,
    CoverageSample,
)
from cg.constants import Workflow
from cg.constants.constants import GenomeVersion
from cg.constants.nf_analysis import (
    RAREDISEASE_COVERAGE_FILE_TAGS,
    RAREDISEASE_COVERAGE_INTERVAL_TYPE,
    RAREDISEASE_COVERAGE_THRESHOLD,
    RAREDISEASE_METRIC_CONDITIONS_WES,
    RAREDISEASE_METRIC_CONDITIONS_WGS,
    RAREDISEASE_PARENT_PEDDY_METRIC_CONDITION,
)
from cg.constants.scout import RAREDISEASE_CASE_TAGS
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase, MultiqcDataJson
from cg.models.raredisease.raredisease import RarediseaseQCMetrics
from cg.resources import RAREDISEASE_BUNDLE_FILENAMES_PATH
from cg.store.models import Sample

LOG = logging.getLogger(__name__)


class RarediseaseAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RAREDISEASE processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.RAREDISEASE,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.raredisease.root
        self.workflow_bin_path: str = config.raredisease.workflow_bin_path
        self.profile: str = config.raredisease.profile
        self.conda_env: str = config.raredisease.conda_env
        self.conda_binary: str = config.raredisease.conda_binary
        self.platform: str = config.raredisease.platform
        self.params: str = config.raredisease.params
        self.workflow_config_path: str = config.raredisease.config
        self.resources: str = config.raredisease.resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.raredisease.tower_workflow
        self.account: str = config.raredisease.slurm.account
        self.email: str = config.raredisease.slurm.mail_user
        self.revision: str = config.raredisease.revision

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Raredisease bundle filenames path."""
        return RAREDISEASE_BUNDLE_FILENAMES_PATH

    def get_workflow_metrics(self, sample_id: str) -> dict:
        """Return Raredisease workflow metric conditions for a sample."""
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=sample_id)
        if "-" not in sample_id:
            metric_conditions: dict[str, dict[str, Any]] = (
                self.get_metric_conditions_by_prep_category(sample_id=sample.internal_id)
            )
            self.set_order_sex_for_sample(sample=sample, metric_conditions=metric_conditions)
        else:
            metric_conditions = RAREDISEASE_PARENT_PEDDY_METRIC_CONDITION.copy()
        return metric_conditions

    def get_metric_conditions_by_prep_category(self, sample_id: str) -> dict:
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=sample_id)
        if (
            sample.application_version.application.analysis_type
            == SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
        ):
            return RAREDISEASE_METRIC_CONDITIONS_WGS.copy()
        return RAREDISEASE_METRIC_CONDITIONS_WES.copy()

    def _get_sample_pair_patterns(self, case_id: str) -> list[str]:
        """Return sample-pair patterns for searching in MultiQC."""
        sample_ids: list[str] = list(self.status_db.get_sample_ids_by_case_id(case_id=case_id))
        pairwise_patterns: list[str] = [
            f"{sample1}-{sample2}" for sample1, sample2 in permutations(sample_ids, 2)
        ]
        return pairwise_patterns

    def get_parent_error_ped_check_metric(
        self, pair_sample_ids: str, multiqc_raw_data: dict[dict]
    ) -> MetricsBase | None:
        """Return the parsed metrics for pedigree error given a concatenated pair of sample ids."""
        metric_name: str = "parent_error_ped_check"
        peddy_metrics: dict[str, dict] = multiqc_raw_data["multiqc_peddy"]
        if sample_pair_metrics := peddy_metrics.get(pair_sample_ids, None):
            return self.get_multiqc_metric(
                metric_name=metric_name,
                metric_value=sample_pair_metrics[metric_name],
                metric_id=pair_sample_ids,
            )

    def get_raredisease_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the metrics specified in a MultiQC json file."""
        multiqc_json: MultiqcDataJson = self.get_multiqc_data_json(case_id=case_id)
        metrics = []
        for search_pattern, metric_id in self.get_multiqc_search_patterns(case_id).items():
            metrics_for_pattern: list[MetricsBase] = (
                self.get_metrics_from_multiqc_json_with_pattern(
                    search_pattern=search_pattern,
                    multiqc_json=multiqc_json,
                    metric_id=metric_id,
                    exact_match=self.is_multiqc_pattern_search_exact,
                )
            )
            metrics.extend(metrics_for_pattern)
        for sample_pair in self._get_sample_pair_patterns(case_id):
            if parent_error_metric := self.get_parent_error_ped_check_metric(
                pair_sample_ids=sample_pair, multiqc_raw_data=multiqc_json.report_saved_raw_data
            ):
                metrics.append(parent_error_metric)
        metrics = self.get_deduplicated_metrics(metrics=metrics)
        return metrics

    def create_metrics_deliverables_content(self, case_id: str) -> dict[str, list[dict[str, Any]]]:
        """Create the content of a Raredisease metrics deliverables file."""
        metrics: list[MetricsBase] = self.get_raredisease_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)
        return {"metrics": [metric.dict() for metric in metrics]}

    @staticmethod
    def set_order_sex_for_sample(sample: Sample, metric_conditions: dict) -> None:
        metric_conditions["predicted_sex_sex_check"]["threshold"] = sample.sex
        metric_conditions["gender"]["threshold"] = sample.sex

    def get_sample_coverage_file_path(self, bundle_name: str, sample_id: str) -> str | None:
        """Return the Raredisease d4 coverage file path."""
        coverage_file_tags: list[str] = RAREDISEASE_COVERAGE_FILE_TAGS + [sample_id]
        coverage_file: File | None = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=bundle_name, tags=coverage_file_tags
        )
        if coverage_file:
            return coverage_file.full_path
        LOG.warning(f"No coverage file found with the tags: {coverage_file_tags}")
        return None

    def get_sample_coverage(
        self, case_id: str, sample_id: str, gene_ids: list[int]
    ) -> CoverageMetrics | None:
        """Return sample coverage metrics from Chanjo2."""
        genome_version: GenomeVersion = self.get_genome_build(case_id)
        coverage_file_path: str | None = self.get_sample_coverage_file_path(
            bundle_name=case_id, sample_id=sample_id
        )
        try:
            post_request = CoveragePostRequest(
                build=self.translate_genome_reference(genome_version),
                coverage_threshold=RAREDISEASE_COVERAGE_THRESHOLD,
                hgnc_gene_ids=gene_ids,
                interval_type=RAREDISEASE_COVERAGE_INTERVAL_TYPE,
                samples=[CoverageSample(coverage_file_path=coverage_file_path, name=sample_id)],
            )
            post_response: CoveragePostResponse = self.chanjo2_api.get_coverage(post_request)
            return post_response.get_sample_coverage_metrics(sample_id)
        except Exception as error:
            LOG.error(f"Error getting coverage for sample '{sample_id}', error: {error}")
            return None

    def get_scout_upload_case_tags(self) -> dict:
        """Return Raredisease Scout upload case tags."""
        return RAREDISEASE_CASE_TAGS

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        """Return reference genome for a raredisease case. Currently fixed for hg19."""
        return GenomeVersion.HG19

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> NextflowAnalysis:
        """Parse Nextflow output analysis files and return an analysis model."""
        qc_metrics_model = RarediseaseQCMetrics
        return super().parse_analysis(
            qc_metrics_raw=qc_metrics_raw, qc_metrics_model=qc_metrics_model, **kwargs
        )



from pydantic import BaseModel, Field, PrivateAttr, model_validator

from cg.models.orders.sample_base import NAME_PATTERN, ContainerEnum, PriorityEnum


class Sample(BaseModel):
    application: str = Field(min_length=1)
    _case_name: str = PrivateAttr(default="")
    _case_priority: PriorityEnum | None = PrivateAttr(default=None)
    comment: str | None = None
    container: ContainerEnum
    container_name: str | None = Field(default=None, pattern=NAME_PATTERN)
    _generated_lims_id: str | None = PrivateAttr(default=None)  # Will be populated by LIMS
    name: str = Field(pattern=NAME_PATTERN, min_length=2, max_length=128)
    volume: int | None = None
    well_position: str | None = None

    @property
    def is_new(self) -> bool:
        return True

    @property
    def is_on_plate(self) -> bool:
        return self.container == ContainerEnum.plate

    @model_validator(mode="before")
    @classmethod
    def convert_empty_strings_to_none(cls, data):
        if isinstance(data, dict):
            for key, value in data.items():
                if value == "":
                    data[key] = None
        return data

    @model_validator(mode="after")
    def set_tube_name_default(self):
        if self.container == ContainerEnum.tube and not self.container_name:
            self.container_name = self.name
        return self



"""Data transfer objects for the PacbioDataTransferService."""

from datetime import datetime

from pydantic import BaseModel

from cg.constants.devices import DeviceType, RevioNames
from cg.services.run_devices.abstract_models import PostProcessingDTOs


class PacBioSMRTCellMetricsDTO(BaseModel):
    type: DeviceType
    well: str
    plate: int
    run_id: str
    movie_name: str
    started_at: datetime
    completed_at: datetime
    hifi_reads: int
    hifi_yield: int
    hifi_mean_read_length: int
    hifi_median_read_length: int
    hifi_mean_length_n50: int
    hifi_median_read_quality: str
    percent_reads_passing_q30: float
    productive_zmws: int
    p0_percent: float
    p1_percent: float
    p2_percent: float
    polymerase_mean_read_length: int
    polymerase_read_length_n50: int
    polymerase_mean_longest_subread: int
    polymerase_longest_subread_n50: int
    control_reads: int
    control_mean_read_length: int
    control_mean_read_concordance: float
    control_mode_read_concordance: float
    failed_reads: int
    failed_yield: int
    failed_mean_read_length: int
    barcoded_hifi_reads: int
    barcoded_hifi_reads_percentage: float
    barcoded_hifi_yield: int
    barcoded_hifi_yield_percentage: float
    barcoded_hifi_mean_read_length: int
    unbarcoded_hifi_reads: int
    unbarcoded_hifi_yield: int
    unbarcoded_hifi_mean_read_length: int


class PacBioSMRTCellDTO(BaseModel):
    type: DeviceType
    internal_id: str


class PacBioSampleSequencingMetricsDTO(BaseModel):
    sample_internal_id: str
    hifi_reads: int
    hifi_yield: int
    hifi_mean_read_length: int
    hifi_median_read_quality: str
    polymerase_mean_read_length: int


class PacBioSequencingRunDTO(BaseModel):
    instrument_name: RevioNames
    run_id: str


class PacBioDTOs(PostProcessingDTOs):
    run_device: PacBioSMRTCellDTO
    smrt_cell_metrics: PacBioSMRTCellMetricsDTO
    sample_sequencing_metrics: list[PacBioSampleSequencingMetricsDTO]
    sequencing_run: PacBioSequencingRunDTO



from pathlib import Path


class PipelineExtension:
    def configure(self, case_id: str, case_run_directory: Path) -> None:
        """Intended for pipeline specific configurations. If none is needed, this bare class
        can be used."""
        pass

    def do_required_files_exist(self, case_run_directory: Path) -> bool:
        return True



from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.exc import OrderNotDeliverableError
from cg.server.dto.delivery_message.delivery_message_request import DeliveryMessageRequest
from cg.server.dto.delivery_message.delivery_message_response import (
    DeliveryMessageOrderResponse,
    DeliveryMessageResponse,
)
from cg.services.delivery_message.utils import get_message, validate_cases
from cg.store.models import Case, Order
from cg.store.store import Store


class DeliveryMessageService:
    def __init__(self, store: Store, trailblazer_api: TrailblazerAPI) -> None:
        self.store = store
        self.trailblazer_api = trailblazer_api

    def get_cases_message(
        self, delivery_message_request: DeliveryMessageRequest
    ) -> DeliveryMessageResponse:
        case_ids: list[str] = delivery_message_request.case_ids
        message: str = self._get_delivery_message(set(case_ids))
        return DeliveryMessageResponse(message=message)

    def get_order_message(self, order_id: int) -> DeliveryMessageOrderResponse:
        order: Order = self.store.get_order_by_id(order_id)
        analyses: list[TrailblazerAnalysis] = self.trailblazer_api.get_analyses_to_deliver(order_id)
        validated_analyses: list[TrailblazerAnalysis] = self._get_validated_analyses(
            analyses=analyses, order=order
        )
        if not validated_analyses:
            raise OrderNotDeliverableError(
                f"No analyses ready to be delivered for order {order_id}"
            )
        case_ids: set[str] = {analysis.case_id for analysis in validated_analyses}
        message: str = self._get_delivery_message(case_ids)
        analysis_ids: list[int] = [analysis.id for analysis in validated_analyses]
        return DeliveryMessageOrderResponse(message=message, analysis_ids=analysis_ids)

    @staticmethod
    def _get_validated_analyses(
        analyses: list[TrailblazerAnalysis], order: Order
    ) -> list[TrailblazerAnalysis]:
        """Returns only the uploaded analyses with workflow matching the order."""
        return [
            analysis
            for analysis in analyses
            if analysis.uploaded_at and analysis.workflow.lower() == order.workflow
        ]

    def _get_delivery_message(self, case_ids: set[str]) -> str:
        cases: list[Case] = self.store.get_cases_by_internal_ids(case_ids)
        validate_cases(cases=cases, case_ids=case_ids)
        message: str = get_message(cases=cases, store=self.store)
        return message



from collections.abc import Callable
from enum import Enum

from cg.constants import Workflow
from cg.services.sequencing_qc_service.quality_checks.utils import (
    all_samples_in_case_have_reads,
    any_sample_in_case_has_reads,
    case_pass_sequencing_qc_on_hifi_yield,
    case_pass_sequencing_qc_on_reads,
    raw_data_case_pass_qc,
    sample_pass_sequencing_qc_on_reads,
)
from cg.store.models import Case


class QualityCheck(Enum):
    """
    Parent class for the quality checks.
    """


class SequencingQCCheck(QualityCheck):
    CASE_PASSES_ON_READS: Callable = case_pass_sequencing_qc_on_reads
    CASE_PASSES_ON_YIELD: Callable = case_pass_sequencing_qc_on_hifi_yield
    SAMPLE_PASSES: Callable = sample_pass_sequencing_qc_on_reads
    ALL_SAMPLES_IN_CASE_HAVE_READS: Callable = all_samples_in_case_have_reads
    ANY_SAMPLE_IN_CASE_HAS_READS: Callable = any_sample_in_case_has_reads
    RAW_DATA_CASE_QC: Callable = raw_data_case_pass_qc


def get_sequencing_quality_check_for_case(case: Case) -> Callable:
    """Return the appropriate sequencing quality checks for the workflow for a case."""
    workflow: Workflow = case.data_analysis

    workflow_qc_mapping = {
        Workflow.BALSAMIC: SequencingQCCheck.CASE_PASSES_ON_READS,
        Workflow.BALSAMIC_PON: SequencingQCCheck.CASE_PASSES_ON_READS,
        Workflow.BALSAMIC_UMI: SequencingQCCheck.CASE_PASSES_ON_READS,
        Workflow.FLUFFY: SequencingQCCheck.ANY_SAMPLE_IN_CASE_HAS_READS,
        Workflow.MICROSALT: SequencingQCCheck.ANY_SAMPLE_IN_CASE_HAS_READS,
        Workflow.MIP_DNA: SequencingQCCheck.CASE_PASSES_ON_READS,
        Workflow.MIP_RNA: SequencingQCCheck.CASE_PASSES_ON_READS,
        Workflow.MUTANT: SequencingQCCheck.ANY_SAMPLE_IN_CASE_HAS_READS,
        Workflow.NALLO: SequencingQCCheck.CASE_PASSES_ON_YIELD,
        Workflow.RAREDISEASE: SequencingQCCheck.CASE_PASSES_ON_READS,
        Workflow.RAW_DATA: SequencingQCCheck.RAW_DATA_CASE_QC,
        Workflow.RNAFUSION: SequencingQCCheck.CASE_PASSES_ON_READS,
        Workflow.TAXPROFILER: SequencingQCCheck.ALL_SAMPLES_IN_CASE_HAVE_READS,
        Workflow.TOMTE: SequencingQCCheck.CASE_PASSES_ON_READS,
    }

    if workflow in workflow_qc_mapping:
        return workflow_qc_mapping[workflow]

    raise ValueError(f"Workflow {workflow} does not have a sequencing quality check.")


def get_sample_sequencing_quality_check() -> Callable:
    return SequencingQCCheck.SAMPLE_PASSES


def run_quality_checks(quality_checks: list[Callable], **kwargs) -> bool:
    return all(quality_check(**kwargs) for quality_check in quality_checks)



from cg.services.orders.validation.order_types.balsamic.validation_rules import (
    BALSAMIC_CASE_RULES,
    BALSAMIC_CASE_SAMPLE_RULES,
)

BALSAMIC_UMI_CASE_RULES: list[callable] = BALSAMIC_CASE_RULES.copy()
BALSAMIC_UMI_CASE_SAMPLE_RULES: list[callable] = BALSAMIC_CASE_SAMPLE_RULES.copy()



"""Enums for run_devices."""

from enum import Enum, StrEnum, auto


class DeviceType(Enum):
    ILLUMINA = auto()
    BIONANO = auto()
    OXFORD_NANOPORE = auto()
    PACBIO = auto()


class RevioNames(StrEnum):
    BETTY = "Betty"
    WILMA = "Wilma"



from cg.constants.constants import Workflow
from cg.constants.priority import SlurmQos
from cg.services.analysis_starter.configurator.abstract_model import CaseConfig


class MIPDNACaseConfig(CaseConfig):
    conda_binary: str
    conda_environment: str
    email: str
    pipeline_binary: str
    pipeline_command: str
    pipeline_config_path: str
    slurm_qos: SlurmQos
    start_after: str | None = None
    start_with: str | None = None
    workflow: Workflow = Workflow.MIP_DNA
    use_bwa_mem: bool

    def get_start_command(self) -> str:
        start_command = (
            "{conda_binary} run --name {conda_environment} {pipeline_binary} {pipeline_command}"
            " --config {pipeline_config_path} {case_id} --slurm_quality_of_service "
            "{slurm_qos} --email {email}"
        ).format(**self.model_dump())

        if self.start_after:
            start_command += f" --start_after_recipe {self.start_after}"

        if self.start_with:
            start_command += f" --start_with_recipe {self.start_with}"

        if self.use_bwa_mem:
            start_command += " --bwa_mem 1 --bwa_mem2 0"

        return start_command



from cg.constants.sample_sources import SourceType
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.models.orders.constants import OrderType
from cg.models.orders.sample_base import ContainerEnum
from cg.services.orders.validation.constants import MAXIMUM_VOLUME, MINIMUM_VOLUME
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.models.sample_aliases import (
    SampleWithCaptureKit,
    SampleWithSkipRC,
)
from cg.store.models import Application
from cg.store.store import Store


def is_volume_invalid(sample: Sample) -> bool:
    in_container: bool = is_in_container(sample.container)
    allowed_volume: bool = is_volume_within_allowed_interval(sample.volume)
    return in_container and not allowed_volume


def is_in_container(container: ContainerEnum) -> bool:
    return container != ContainerEnum.no_container


def is_volume_within_allowed_interval(volume: int) -> bool:
    return volume and (MINIMUM_VOLUME <= volume <= MAXIMUM_VOLUME)


def is_sample_on_plate(sample: Sample) -> bool:
    return sample.container == ContainerEnum.plate


def is_application_compatible(
    order_type: OrderType,
    application_tag: str,
    store: Store,
) -> bool:
    application: Application | None = store.get_application_by_tag(application_tag)
    return not application or order_type in application.order_types


def is_volume_missing(sample: Sample) -> bool:
    """Check if a sample is missing its volume."""
    if is_in_container(sample.container) and not sample.volume:
        return True
    return False


def has_sample_invalid_concentration(
    sample: SampleWithSkipRC, allowed_interval: tuple[float, float]
) -> bool:
    concentration: float | None = sample.concentration_ng_ul
    return concentration and not is_sample_concentration_within_interval(
        concentration=concentration, interval=allowed_interval
    )


def get_concentration_interval(
    sample: SampleWithSkipRC, application: Application
) -> tuple[float, float] | None:
    is_cfdna: bool = is_sample_cfdna(sample)
    allowed_interval: tuple[float, float] = get_application_concentration_interval(
        application=application, is_cfdna=is_cfdna
    )
    return allowed_interval


def is_sample_cfdna(sample: SampleWithSkipRC) -> bool:
    source = sample.source
    return source == SourceType.CELL_FREE_DNA


def get_application_concentration_interval(
    application: Application, is_cfdna: bool
) -> tuple[float, float]:
    if is_cfdna:
        return (
            application.sample_concentration_minimum_cfdna,
            application.sample_concentration_maximum_cfdna,
        )
    return application.sample_concentration_minimum, application.sample_concentration_maximum


def is_sample_concentration_within_interval(
    concentration: float, interval: tuple[float, float]
) -> bool:
    return interval[0] <= concentration <= interval[1]


def does_sample_need_capture_kit(sample: SampleWithCaptureKit, store: Store) -> bool:
    application: Application | None = store.get_application_by_tag(sample.application)
    return (
        application
        and application.prep_category == SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING
    )


def is_sample_missing_capture_kit(sample: SampleWithCaptureKit, store: Store) -> bool:
    """Returns whether a TGS sample has an application and is missing a capture kit."""
    return does_sample_need_capture_kit(sample=sample, store=store) and not sample.capture_kit


def is_invalid_capture_kit(sample: SampleWithCaptureKit, store: Store) -> bool:
    if not sample.capture_kit:
        return False
    valid_beds: list[str] = [bed.name for bed in store.get_active_beds()]
    return sample.capture_kit not in valid_beds



from pydantic import BaseModel, Field

from cg.models.orders.sample_base import NAME_PATTERN, StatusEnum


class ExistingSample(BaseModel):
    father: str | None = Field(None, pattern=NAME_PATTERN)
    internal_id: str
    mother: str | None = Field(None, pattern=NAME_PATTERN)
    status: StatusEnum | None = None

    @property
    def is_new(self) -> bool:
        return False



from enum import StrEnum

from cg.constants import DataDelivery


class MutantDeliveryType(StrEnum):
    FASTQ_ANALYSIS = DataDelivery.FASTQ_ANALYSIS
    NO_DELIVERY = DataDelivery.NO_DELIVERY


class PreProcessingMethod(StrEnum):
    COVID_PRIMER = "Qiagen SARS-CoV-2 Primer Panel"
    COVID_SEQUENCING = "COVIDSeq"
    OTHER = 'Other (specify in "Comments")'


class Primer(StrEnum):
    ILLUMINA = "Illumina Artic V3"
    NANOPORE = "Nanopore Midnight V1"


class Region(StrEnum):
    STOCKHOLM = "Stockholm"
    UPPSALA = "Uppsala"
    SORMLAND = "Sörmland"
    OSTERGOTLAND = "Östergötland"
    JONKOPINGS_LAN = "Jönköpings län"
    KRONOBERG = "Kronoberg"
    KALMAR_LAN = "Kalmar län"
    GOTLAND = "Gotland"
    BLEKINGE = "Blekinge"
    SKANE = "Skåne"
    HALLAND = "Halland"
    VASTRA_GOTALANDSREGIONEN = "Västra Götalandsregionen"
    VARMLAND = "Värmland"
    OREBRO_LAN = "Örebro län"
    VASTMANLAND = "Västmanland"
    DALARNA = "Dalarna"
    GAVLEBORG = "Gävleborg"
    VASTERNORRLAND = "Västernorrland"
    JAMTLAND_HARJEDALEN = "Jämtland Härjedalen"
    VASTERBOTTEN = "Västerbotten"
    NORRBOTTEN = "Norrbotten"


class SelectionCriteria(StrEnum):
    ALLMAN_OVERVAKNING = "Allmän övervakning"
    ALLMAN_OVERVAKNING_OPPENVARD = "Allmän övervakning öppenvård"
    ALLMAN_OVERVAKNING_SLUTENVARD = "Allmän övervakning slutenvård"
    UTLANDSVISTELSE = "Utlandsvistelse"
    RIKTAD_INSAMLING = "Riktad insamling"
    UTBROTT = "Utbrott"
    VACCINATIONSGENOMBROTT = "Vaccinationsgenombrott"
    REINFEKTION = "Reinfektion"
    INFORMATION_SAKNAS = "Information saknas"


class OriginalLab(StrEnum):
    UNILABS_STOCKHOLM = "Unilabs Stockholm"
    UNILABS_ESKILSTUNA_LABORATORIUM = "Unilabs Eskilstuna Laboratorium"
    NORRLAND_UNIVERSITY_HOSPITAL = "Norrland University Hospital"
    LANSSJUKHUSET_SUNDSVALL = "Länssjukhuset Sundsvall"
    A05_DIAGNOSTICS = "A05 Diagnostics"
    SYNLAB_MEDILAB = "Synlab Medilab"
    KAROLINSKA_UNIVERSITY_HOSPITAL_SOLNA = "Karolinska University Hospital Solna"
    KAROLINSKA_UNIVERSITY_HOSPITAL_HUDDINGE = "Karolinska University Hospital Huddinge"
    LABORATORIEMEDICINSKT_CENTRUM_GOTLAND = "LaboratorieMedicinskt Centrum Gotland"



from cg.apps.tb.dto.summary_response import AnalysisSummary


def _get_analysis_map(analysis_summaries: list[AnalysisSummary]) -> dict:
    return {summary.order_id: summary for summary in analysis_summaries}



import logging
from pathlib import Path
from typing import Any

from pydantic import ValidationError

from cg.apps.mip.confighandler import ConfigHandler
from cg.constants import FileExtensions, Workflow
from cg.constants.constants import FileFormat
from cg.constants.housekeeper_tags import HkMipAnalysisTag
from cg.exc import CgError
from cg.io.controller import ReadFile, WriteFile
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import MipFastqHandler
from cg.models.cg_config import CGConfig
from cg.models.mip.mip_analysis import MipAnalysis
from cg.models.mip.mip_config import MipBaseConfig
from cg.models.mip.mip_metrics_deliverables import MIPMetricsDeliverables
from cg.models.mip.mip_sample_info import MipBaseSampleInfo
from cg.store.models import BedVersion, Case, CaseSample, Sample

CLI_OPTIONS = {
    "config": {"option": "--config_file"},
    "slurm_quality_of_service": {"option": "--slurm_quality_of_service"},
    "email": {"option": "--email"},
    "base": {"option": "--cluster_constant_path"},
    "dryrun": {"option": "--dry_run_all"},
    "gene_list": {"option": "--vcfparser_slt_fl"},
    "max_gaussian": {"option": "--gatk_varrecal_snv_max_gau"},
    "skip_evaluation": {"option": "--qccollect_skip_evaluation"},
    "start_after": {"option": "--start_after_recipe"},
    "start_with": {"option": "--start_with_recipe"},
    "use_bwa_mem": {"option": "--bwa_mem 1 --bwa_mem2 0"},
}
LOG = logging.getLogger(__name__)


class MipAnalysisAPI(AnalysisAPI):
    """The workflow is accessed through Trailblazer but cg provides additional conventions and
    hooks into the status database that makes managing analyses simpler"""

    def __init__(self, config: CGConfig, workflow: Workflow):
        super().__init__(workflow, config)

    @property
    def root(self) -> str:
        raise NotImplementedError

    @property
    def conda_env(self) -> str:
        raise NotImplementedError

    @property
    def mip_workflow(self) -> str:
        raise NotImplementedError

    @property
    def script(self) -> str:
        raise NotImplementedError

    @property
    def fastq_handler(self):
        return MipFastqHandler

    def get_pedigree_config_path(self, case_id: str) -> Path:
        return Path(self.root, case_id, "pedigree.yaml")

    def get_case_config_path(self, case_id: str) -> Path:
        return Path(self.root, case_id, "analysis", f"{case_id}_config.yaml")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """
        Location in working directory where deliverables file will be stored upon completion of analysis.
        Deliverables file is used to communicate paths and tag definitions for files in a finished analysis
        """
        return Path(self.root, case_id, "analysis", f"{case_id}_deliverables.yaml")

    def get_sample_info_path(self, case_id: str) -> Path:
        """Get case analysis sample info path"""
        return Path(self.root, case_id, "analysis", f"{case_id}_qc_sample_info.yaml")

    def get_panel_bed(self, panel_bed: str = None) -> str | None:
        """Check and return BED gene panel."""
        if not panel_bed:
            return None
        if panel_bed.endswith(FileExtensions.BED):
            return panel_bed
        bed_version: BedVersion | None = self.status_db.get_bed_version_by_short_name(
            bed_version_short_name=panel_bed
        )
        if not bed_version:
            raise CgError("Please provide a valid panel shortname or a path to panel.bed file!")
        return bed_version.filename

    def pedigree_config(self, case_id: str, panel_bed: str = None) -> dict:
        """Make the MIP pedigree config. Meta data for the family is taken from the family object
        and converted to MIP format via trailblazer.
        """

        # Validate and reformat to MIP pedigree config format
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        return ConfigHandler.make_pedigree_config(
            data={
                "case": case_obj.internal_id,
                "default_gene_panels": case_obj.panels,
                "samples": [
                    self.config_sample(link_obj=link_obj, panel_bed=panel_bed)
                    for link_obj in case_obj.links
                ],
            }
        )

    def write_pedigree_config(self, case_id: str, data: dict) -> None:
        """Write the pedigree config to the the case dir"""
        out_dir = self.get_case_path(case_id=case_id)
        out_dir.mkdir(parents=True, exist_ok=True)
        pedigree_config_path: Path = self.get_pedigree_config_path(case_id=case_id)
        WriteFile.write_file_from_content(
            content=data, file_format=FileFormat.YAML, file_path=pedigree_config_path
        )
        LOG.info("Config file saved to %s", pedigree_config_path)

    @staticmethod
    def get_sample_data(link_obj: CaseSample) -> dict[str, str | int]:
        """Return sample specific data."""
        return {
            "sample_id": link_obj.sample.internal_id,
            "sample_display_name": link_obj.sample.name,
            "analysis_type": link_obj.sample.application_version.application.analysis_type,
            "sex": link_obj.sample.sex,
            "phenotype": link_obj.status,
            "expected_coverage": link_obj.sample.application_version.application.min_sequencing_depth,
        }

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        """Return the path to the FASTQ destination directory."""
        return Path(
            self.root,
            case.internal_id,
            sample.application_version.application.analysis_type,
            sample.internal_id,
            FileFormat.FASTQ,
        )

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        for link in case.links:
            self.link_fastq_files_for_sample(case=case, sample=link.sample)

    def _get_latest_raw_file(self, family_id: str, tags: list[str]) -> Any:
        """Get a python object file for a tag and a family ."""

        last_version = self.housekeeper_api.last_version(bundle=family_id)

        analysis_files = self.housekeeper_api.files(
            bundle=family_id, version=last_version.id, tags=tags
        ).all()

        if analysis_files:
            analysis_file_raw = self._open_bundle_file(analysis_files[0].path)
        else:
            raise LOG.warning(
                "No post analysis files received from DeliverAPI for '%s'",
                family_id,
            )

        return analysis_file_raw

    def _open_bundle_file(self, relative_file_path: str) -> Any:
        """Open a bundle file and return it as an Python object"""

        full_file_path: Path = Path(self.housekeeper_api.get_root_dir()).joinpath(
            relative_file_path
        )
        return ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=full_file_path)

    def get_latest_metadata(self, family_id: str) -> MipAnalysis:
        """Return the latest trending data for a family."""

        mip_config_raw = self._get_latest_raw_file(
            family_id=family_id, tags=HkMipAnalysisTag.CONFIG
        )
        qc_metrics_raw = self._get_latest_raw_file(
            family_id=family_id, tags=HkMipAnalysisTag.QC_METRICS
        )
        sample_info_raw = self._get_latest_raw_file(
            family_id=family_id, tags=HkMipAnalysisTag.SAMPLE_INFO
        )
        if mip_config_raw and qc_metrics_raw and sample_info_raw:
            try:
                mip_analysis: MipAnalysis = self.parse_analysis(
                    config_raw=mip_config_raw,
                    qc_metrics_raw=qc_metrics_raw,
                    sample_info_raw=sample_info_raw,
                )
                return mip_analysis
            except ValidationError as error:
                LOG.error(
                    "get_latest_metadata failed for '%s', missing attribute: %s",
                    family_id,
                    error,
                )
                raise error
        else:
            LOG.error(f"Unable to retrieve the latest metadata for {family_id}")
            raise CgError

    def parse_analysis(
        self, config_raw: dict, qc_metrics_raw: dict, sample_info_raw: dict
    ) -> MipAnalysis:
        """Parses the output analysis files from MIP

        Args:
            config_raw (dict): raw YAML input from MIP analysis config file
            qc_metrics_raw (dict): raw YAML input from MIP analysis qc metric file
            sample_info_raw (dict): raw YAML input from MIP analysis qc sample info file
        Returns:
            MipAnalysis: parsed MIP analysis data
        """
        mip_config: MipBaseConfig = MipBaseConfig(**config_raw)
        qc_metrics = MIPMetricsDeliverables(**qc_metrics_raw)
        sample_info: MipBaseSampleInfo = MipBaseSampleInfo(**sample_info_raw)

        return MipAnalysis(
            case=mip_config.case_id,
            genome_build=sample_info.genome_build,
            sample_id_metrics=qc_metrics.sample_id_metrics,
            mip_version=sample_info.mip_version,
            rank_model_version=sample_info.rank_model_version,
            sample_ids=mip_config.sample_ids,
            sv_rank_model_version=sample_info.sv_rank_model_version,
        )

    def get_skip_evaluation_flag(self, case_id: str, skip_evaluation: bool) -> bool:
        """If any sample in this case is downsampled or external, returns true"""
        if skip_evaluation:
            return True
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        for link_obj in case_obj.links:
            downsampled = isinstance(link_obj.sample.downsampled_to, int)
            external = link_obj.sample.application_version.application.is_external
            if downsampled or external:
                LOG.info(
                    "%s: downsampled/external - skip evaluation",
                    link_obj.sample.internal_id,
                )
                return True
        return False

    @staticmethod
    def _append_value_for_non_flags(parameters: list, value) -> None:
        """Add the value of the non boolean options to the parameters"""
        if value is not True:
            parameters.append(value)

    @staticmethod
    def _cg_to_mip_option_map(parameters: list, mip_key) -> None:
        """Map cg options to MIP option syntax"""
        parameters.append(CLI_OPTIONS[mip_key]["option"])

    def run_analysis(self, case_id: str, command_args: dict, dry_run: bool) -> None:
        parameters = [
            case_id,
        ]
        for key, value in command_args.items():
            if value:
                self._cg_to_mip_option_map(parameters, key)
                self._append_value_for_non_flags(parameters, value)

        exit_code = self.process.run_command(dry_run=dry_run, parameters=parameters)
        for line in self.process.stdout_lines():
            LOG.info(line)
        for line in self.process.stderr_lines():
            LOG.info(line)
        return exit_code

    def get_case_path(self, case_id: str) -> Path:
        return Path(self.root, case_id)

    def get_job_ids_path(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "analysis", "slurm_job_ids.yaml")

    def config_sample(self, link_obj: CaseSample, panel_bed: str) -> dict:
        raise NotImplementedError

    def get_workflow_version(self, case_id: str) -> str:
        """Get MIP version from sample info file"""
        LOG.debug("Fetch workflow version")
        sample_info_raw: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=self.get_sample_info_path(case_id)
        )
        sample_info: MipBaseSampleInfo = MipBaseSampleInfo(**sample_info_raw)
        return sample_info.mip_version

    def write_managed_variants(self, case_id: str, content: list[str]) -> None:
        self._write_managed_variants(out_dir=Path(self.root, case_id), content=content)



from cg.services.orders.validation.order_types.fluffy.models.order import FluffyOrder
from cg.services.orders.validation.order_types.metagenome.models.sample import MetagenomeSample
from cg.services.orders.validation.order_types.microbial_fastq.models.order import (
    MicrobialFastqOrder,
)
from cg.services.orders.validation.order_types.microsalt.models.order import MicrosaltOrder
from cg.services.orders.validation.order_types.mutant.models.order import MutantOrder
from cg.services.orders.validation.order_types.rml.models.order import RMLOrder
from cg.services.orders.validation.order_types.taxprofiler.models.sample import TaxprofilerSample

OrderWithIndexedSamples = FluffyOrder | RMLOrder
OrderWithControlSamples = (
    MetagenomeSample | MicrobialFastqOrder | MicrosaltOrder | MutantOrder | TaxprofilerSample
)



from cg.constants import Workflow
from cg.models.orders.constants import OrderType

ORDER_TYPE_WORKFLOW_MAP: dict[OrderType, Workflow] = {
    OrderType.BALSAMIC: Workflow.BALSAMIC,
    OrderType.BALSAMIC_UMI: Workflow.BALSAMIC_UMI,
    OrderType.FASTQ: Workflow.RAW_DATA,
    OrderType.FLUFFY: Workflow.FLUFFY,
    OrderType.METAGENOME: Workflow.RAW_DATA,
    OrderType.MICROBIAL_FASTQ: Workflow.RAW_DATA,
    OrderType.MICROSALT: Workflow.MICROSALT,
    OrderType.MIP_DNA: Workflow.MIP_DNA,
    OrderType.MIP_RNA: Workflow.MIP_RNA,
    OrderType.NALLO: Workflow.NALLO,
    OrderType.PACBIO_LONG_READ: Workflow.RAW_DATA,
    OrderType.RAREDISEASE: Workflow.RAREDISEASE,
    OrderType.RML: Workflow.RAW_DATA,
    OrderType.RNAFUSION: Workflow.RNAFUSION,
    OrderType.SARS_COV_2: Workflow.MUTANT,
    OrderType.TAXPROFILER: Workflow.TAXPROFILER,
    OrderType.TOMTE: Workflow.TOMTE,
}



from enum import StrEnum

from cg.constants.constants import DataDelivery


class RNAFusionDeliveryType(StrEnum):
    ANALYSIS_FILES = DataDelivery.ANALYSIS_FILES
    ANALYSIS_SCOUT = DataDelivery.ANALYSIS_SCOUT
    SCOUT = DataDelivery.SCOUT
    FASTQ = DataDelivery.FASTQ
    FASTQ_ANALYSIS = DataDelivery.FASTQ_ANALYSIS
    FASTQ_SCOUT = DataDelivery.FASTQ_SCOUT
    FASTQ_ANALYSIS_SCOUT = DataDelivery.FASTQ_ANALYSIS_SCOUT
    NO_DELIVERY = DataDelivery.NO_DELIVERY



import copy
import re


def replace_values_in_params_file(workflow_parameters: dict) -> dict:
    """
    Iterate through the dictionary until all placeholders are replaced with the corresponding value
    from the dictionary
    """
    replaced_workflow_parameters = copy.deepcopy(workflow_parameters)
    while True:
        resolved: bool = True
        for key, value in replaced_workflow_parameters.items():
            new_value: str | int = _replace_params_placeholders(value, workflow_parameters)
            if new_value != value:
                resolved = False
                replaced_workflow_parameters[key] = new_value
        if resolved:
            break
    return replaced_workflow_parameters


def _replace_params_placeholders(value: str | int, workflow_parameters: dict) -> str:
    """Replace values marked as placeholders with values from the given dictionary"""
    if isinstance(value, str):
        placeholders: list[str] = re.findall(r"{{\s*([^{}\s]+)\s*}}", value)
        for placeholder in placeholders:
            if placeholder in workflow_parameters:
                value = value.replace(
                    f"{{{{{placeholder}}}}}", str(workflow_parameters[placeholder])
                )
    return value
